# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Support for the gRPC-based Ledger API.
"""
from asyncio import gather
from datetime import datetime
from threading import Event
from time import sleep
from typing import TYPE_CHECKING, AbstractSet, Iterable, Mapping, Optional, Sequence, cast
import warnings

from grpc import (
    Channel,
    RpcError,
    StatusCode,
    insecure_channel,
    secure_channel,
    ssl_channel_credentials,
)

from ... import LOG
from ..._gen.com.daml.ledger.api.v1.active_contracts_service_pb2_grpc import (
    ActiveContractsServiceStub as G_ActiveContractsServiceStub,
)
from ..._gen.com.daml.ledger.api.v1.admin.package_management_service_pb2 import (
    UploadDarFileRequest as G_UploadDarFileRequest,
)
from ..._gen.com.daml.ledger.api.v1.admin.package_management_service_pb2_grpc import (
    PackageManagementServiceStub as G_PackageManagementServiceStub,
)
from ..._gen.com.daml.ledger.api.v1.command_service_pb2_grpc import (
    CommandServiceStub as G_CommandServiceStub,
)
from ..._gen.com.daml.ledger.api.v1.ledger_identity_service_pb2 import (
    GetLedgerIdentityRequest as G_GetLedgerIdentityRequest,
)
from ..._gen.com.daml.ledger.api.v1.ledger_identity_service_pb2_grpc import (
    LedgerIdentityServiceStub as G_LedgerIdentityServiceStub,
)
from ..._gen.com.daml.ledger.api.v1.package_service_pb2 import (
    GetPackageRequest as G_GetPackageRequest,
    ListPackagesRequest as G_ListPackagesRequest,
)
from ..._gen.com.daml.ledger.api.v1.package_service_pb2_grpc import (
    PackageServiceStub as G_PackageServiceStub,
)
from ..._gen.com.daml.ledger.api.v1.testing.time_service_pb2 import (
    GetTimeRequest as G_GetTimeRequest,
    SetTimeRequest as G_SetTimeRequest,
)
from ..._gen.com.daml.ledger.api.v1.testing.time_service_pb2_grpc import (
    TimeServiceStub as G_TimeServiceStub,
)
from ..._gen.com.daml.ledger.api.v1.transaction_pb2 import (
    Transaction as G_Transaction,
    TransactionTree as G_TransactionTree,
)
from ..._gen.com.daml.ledger.api.v1.transaction_service_pb2 import (
    GetLedgerEndRequest as G_GetLedgerEndRequest,
    GetTransactionTreesResponse as G_GetTransactionTreesResponse,
)
from ..._gen.com.daml.ledger.api.v1.transaction_service_pb2_grpc import (
    TransactionServiceStub as G_TransactionServiceStub,
)
from ...damlast.daml_lf_1 import PackageRef
from ...damlast.parse import parse_archive_payload
from ...ledger.aio.pkgloader_compat import PackageLoader
from ...prim import Party, datetime_to_timestamp, to_party
from ...scheduler import Invoker, RunLevel
from ...util.io import read_file_bytes
from ...util.typing import safe_cast
from .._base import LedgerClient, LedgerConnectionOptions, _LedgerConnection
from ..errors import ConnectionTimeoutError, UserTerminateRequest
from .pb_parse_event import (
    BaseEventDeserializationContext,
    serialize_acs_request,
    serialize_transactions_request,
    to_acs_events,
    to_transaction_events,
)

with warnings.catch_warnings():
    warnings.simplefilter("ignore", DeprecationWarning)

    if TYPE_CHECKING:
        from ...client.commands import CommandPayload
        from ...client.ledger import LedgerMetadata
        from ...model.network import HTTPConnectionSettings
        from ...model.reading import BaseEvent, ContractFilter, TransactionFilter
        from ...model.types_store import PackageProvider, PackageStore


__all__ = [
    "GRPCv1LedgerClient",
    "grpc_set_time",
    "grpc_upload_package",
    "grpc_detect_ledger_id",
    "grpc_main_thread",
    "GRPCPackageProvider",
    "grpc_package_sync",
    "grpc_create_channel",
    "GRPCv1Connection",
]


class GRPCv1LedgerClient(LedgerClient):
    def __init__(self, connection: "GRPCv1Connection", ledger: "LedgerMetadata", party: Party):
        self.connection = safe_cast(GRPCv1Connection, connection)
        self.ledger = ledger
        self.party = to_party(party)

    async def commands(self, command_payload: "CommandPayload") -> None:
        from .pb_ser_command import ProtobufSerializer

        serializer = cast(ProtobufSerializer, self.ledger.serializer)
        request = serializer.serialize_command_request(command_payload)
        await self.connection.invoker.run_in_executor(
            lambda: self.connection.command_service.SubmitAndWait(request)
        )
        return None

    async def commands_transaction(self, __1: "CommandPayload") -> G_Transaction:
        from .pb_ser_command import ProtobufSerializer

        serializer = cast(ProtobufSerializer, self.ledger.serializer)
        request = serializer.serialize_command_request(__1)
        response = await self.connection.invoker.run_in_executor(
            lambda: self.connection.command_service.SubmitAndWaitForTransaction(request)
        )
        return response.transaction

    async def commands_transaction_tree(self, __1: "CommandPayload") -> G_TransactionTree:
        from .pb_ser_command import ProtobufSerializer

        serializer = cast(ProtobufSerializer, self.ledger.serializer)
        request = serializer.serialize_command_request(__1)
        response = await self.connection.invoker.run_in_executor(
            lambda: self.connection.command_service.SubmitAndWaitForTransactionTree(request)
        )
        return response.transaction

    async def active_contracts(self, contract_filter: "ContractFilter") -> "Sequence[BaseEvent]":
        with LOG.info_timed("ACS request serialization"):
            request = serialize_acs_request(contract_filter, self.ledger.ledger_id, self.party)
            context = BaseEventDeserializationContext(
                None,
                self.connection.options.lookup,
                self.ledger._store,
                self.party,
                self.ledger.ledger_id,
            )

        with LOG.info_timed("ACS request initial stream"):
            acs_stream = await self.connection.invoker.run_in_executor(
                lambda: self.connection.active_contract_set_service.GetActiveContracts(request)
            )

        with LOG.info_timed("ACS request consume full stream"):
            # Consume a stream of events from the Active Contract Set service and store these results fully
            # in memory; return the full set of events as well as a set of package IDs that need to be
            # available in order to fully understand the contents of the Active Contract Set service.
            acs_responses = list(acs_stream)

        with LOG.info_timed("ACS find all required packages"):
            pkg_refs = {
                create_evt.template_id.package_id
                for acs_response_pb in acs_responses
                for create_evt in acs_response_pb.active_contracts
            }

        if pkg_refs:
            with LOG.info_timed(f"ACS load {len(pkg_refs)} new package(s)"):
                # Preload packages that are DIRECTLY referenced by packages in the message
                await gather(*(self.ledger.package_loader.load(pkg_ref) for pkg_ref in pkg_refs))

        with LOG.info_timed("ACS transform the message"):
            # Now load all the events. Note that do_with_retry is still required because the
            # packages that contained templates might have references to other packages that are
            # also required to understand the individual fields in a template.
            return await self.ledger.package_loader.do_with_retry(
                lambda: to_acs_events(context, acs_responses)
            )

    async def events(self, transaction_filter: "TransactionFilter") -> "Sequence[BaseEvent]":
        request = serialize_transactions_request(
            transaction_filter, self.ledger.ledger_id, self.party
        )
        context = BaseEventDeserializationContext(
            None,
            self.connection.options.lookup,
            self.ledger._store,
            self.party,
            self.ledger.ledger_id,
        )

        tx_future = self.connection.invoker.run_in_executor(
            lambda: self.connection.transaction_service.GetTransactions(request)
        )

        tt_stream = None  # type: Optional[Iterable[G_GetTransactionTreesResponse]]
        if transaction_filter.templates is None:
            # Filtering by package must disable the ability to handle exercise nodes; we may want to
            # consider dropping client-side support for exercise events anyway because they are not
            # widely used
            tt_stream = await self.connection.invoker.run_in_executor(
                lambda: self.connection.transaction_service.GetTransactionTrees(request)
            )

        tx_stream = await tx_future

        return await self.ledger.package_loader.do_with_retry(
            lambda: to_transaction_events(
                context, tx_stream, tt_stream, transaction_filter.destination_offset
            )
        )

    async def events_end(self) -> str:
        request = G_GetLedgerEndRequest(ledger_id=self.ledger.ledger_id)
        return await self.connection.invoker.run_in_executor(
            lambda: self.connection.transaction_service.GetLedgerEnd(request).offset.absolute
        )


def grpc_set_time(connection: "GRPCv1Connection", ledger_id: str, new_datetime: datetime) -> None:
    request = G_GetTimeRequest(ledger_id=ledger_id)
    response = connection.time_service.GetTime(request)
    ts = next(iter(response))

    request = G_SetTimeRequest(
        ledger_id=ledger_id,
        current_time=ts.current_time,
        new_time=datetime_to_timestamp(new_datetime),
    )
    connection.time_service.SetTime(request)
    LOG.info("Time on the server changed by the local client to %s.", new_datetime)


def grpc_upload_package(connection: "GRPCv1Connection", dar_contents: bytes) -> None:
    request = G_UploadDarFileRequest(dar_file=dar_contents)
    connection.package_management_service.UploadDarFile(request)


GRPC_KNOWN_RETRYABLE_ERRORS = (
    "DNS resolution failed",
    "failed to connect to all addresses",
    "no healthy upstream",
)


def grpc_detect_ledger_id(connection: "GRPCv1Connection") -> str:
    """
    Return the ledger ID from the remote server when it becomes available. This method blocks until
    a ledger ID has been successfully retrieved, or the timeout is reached (in which case an
    exception is thrown).
    """
    LOG.debug("Starting a monitor thread for connection: %s", connection)

    start_time = datetime.utcnow()
    connect_timeout = connection.options.connect_timeout

    while connect_timeout is None or (datetime.utcnow() - start_time) < connect_timeout:
        if connection.invoker.level >= RunLevel.TERMINATE_GRACEFULLY:
            raise UserTerminateRequest()
        if connection.closed:
            raise Exception("connection closed")

        try:
            response = connection.ledger_identity_service.GetLedgerIdentity(
                G_GetLedgerIdentityRequest()
            )
        except RpcError as ex:
            details_str = ex.details()

            # suppress some warning strings because they're not particularly useful and just clutter
            # up the logs
            if details_str not in GRPC_KNOWN_RETRYABLE_ERRORS:
                LOG.exception(
                    "An unexpected error occurred when trying to fetch the "
                    "ledger identity; this will be retried."
                )
            sleep(1)
            continue

        return response.ledger_id

    raise ConnectionTimeoutError(
        f"connection timeout exceeded: {connect_timeout.total_seconds()} seconds"
    )


# noinspection PyDeprecation
def grpc_main_thread(connection: "GRPCv1Connection", ledger_id: str) -> "Iterable[LedgerMetadata]":
    from ...client.ledger import LedgerMetadata
    from .pb_ser_command import ProtobufSerializer

    LOG.info("grpc_main_thread...")

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        from ...model.types_store import PackageStore

        store = PackageStore.empty()

        package_provider = GRPCPackageProvider(connection, ledger_id)

        if connection.options.eager_package_fetch:
            grpc_package_sync(package_provider, store)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        package_loader = PackageLoader(
            package_lookup=connection.options.lookup,
            conn=package_provider,
            timeout=connection.options.connect_timeout,
        )

    yield LedgerMetadata(
        ledger_id=ledger_id,
        store=store,
        package_loader=package_loader,
        serializer=ProtobufSerializer(connection.options.lookup),
        protocol_version="v1",
    )

    # poll for package updates once a second
    while not connection._closed.wait(1):
        try:
            if connection.options.eager_package_fetch:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", DeprecationWarning)
                    grpc_package_sync(package_provider, store)
        except Exception as ex:
            if not isinstance(ex, RpcError) or ex.code() != StatusCode.CANCELLED:
                LOG.warning("Package syncing raised an exception.", exc_info=True)

    LOG.debug("The gRPC monitor thread is now shutting down.")


class GRPCPackageProvider:
    def __init__(self, connection: "GRPCv1Connection", ledger_id: str):
        self.connection = connection
        self.ledger_id = ledger_id

    def package_ids(self) -> "AbstractSet[PackageRef]":
        request = G_ListPackagesRequest(ledger_id=self.ledger_id)
        response = self.connection.package_service.ListPackages(request)
        return frozenset(response.package_ids)

    def package_bytes(self, package_id: "PackageRef") -> bytes:
        request = G_GetPackageRequest(ledger_id=self.ledger_id, package_id=package_id)
        package_info = self.connection.package_service.GetPackage(request)
        return package_info.archive_payload

    def get_package_ids(self) -> "AbstractSet[PackageRef]":
        return self.package_ids()

    def fetch_package(self, package_id: "PackageRef") -> bytes:
        return self.package_bytes(package_id)

    def get_all_packages(self) -> "Mapping[PackageRef, bytes]":
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            return {pkg_id: self.fetch_package(pkg_id) for pkg_id in self.get_package_ids()}


# noinspection PyDeprecation
def grpc_package_sync(package_provider: "PackageProvider", store: "PackageStore") -> "None":
    warnings.warn(
        "grpc_package_sync is deprecated; there is no replacement", DeprecationWarning, stacklevel=2
    )

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        from .pb_parse_metadata import find_dependencies, parse_daml_metadata_pb

        LOG.debug("grpc_package_sync started...")

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        all_package_ids = package_provider.get_package_ids()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            loaded_package_ids = {a.hash for a in store.archives()}
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            expected_package_ids = store.expected_package_ids()

        def should_load(p: str) -> bool:
            # TODO: Filtering by expected package IDs may cause packages to never fully load due to
            #  transitive dependencies; here we put the onus on the app writer to ensure that they
            #  supply the complete graph, and we don't even warn them if there is an issue (but
            #  we could only actually warn them if we parse the archive, which is the expensive
            #  operation we're trying to avoid).
            return (
                expected_package_ids is None or p in expected_package_ids
            ) and p not in loaded_package_ids

        metadatas_pb = {}
        for package_id in all_package_ids:
            if should_load(package_id):
                archive_payload = package_provider.fetch_package(package_id)
                metadatas_pb[package_id] = parse_archive_payload(package_id, archive_payload)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        adr = find_dependencies(metadatas_pb, loaded_package_ids)

    for package_id, archive_bytes in adr.sorted_archives.items():
        m = parse_daml_metadata_pb(package_id, archive_bytes)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            store.register_all(m)

    LOG.debug("grpc_package_sync ended.")


def grpc_create_channel(settings: "HTTPConnectionSettings") -> Channel:
    target = f"{settings.host}:{settings.port}"
    options = [("grpc.max_send_message_length", -1), ("grpc.max_receive_message_length", -1)]

    if not settings.enable_http_proxy:
        options.append(("grpc.enable_http_proxy", 0))

    if settings.oauth:
        from google.auth.transport.grpc import secure_authorized_channel
        from google.auth.transport.requests import Request as RefreshRequester
        from google.oauth2.credentials import Credentials as OAuthCredentials

        LOG.debug("Using a secure gRPC connection over OAuth:")

        credentials = OAuthCredentials(
            token=settings.oauth.token,
            refresh_token=settings.oauth.refresh_token,
            id_token=settings.oauth.id_token,
            token_uri=settings.oauth.token_uri,
            client_id=settings.oauth.client_id,
            client_secret=settings.oauth.client_secret,
        )

        ssl_credentials = None
        if settings.ssl_settings:
            cert_chain = read_file_bytes(settings.ssl_settings.cert_file)
            cert = read_file_bytes(settings.ssl_settings.cert_key_file)
            ca_cert = read_file_bytes(settings.ssl_settings.ca_file)

            LOG.debug("Using a secure gRPC connection:")
            LOG.debug("    target: %s", target)
            LOG.debug("    root_certificates: contents of %s", settings.ssl_settings.ca_file)
            LOG.debug("    private_key: contents of %s", settings.ssl_settings.cert_key_file)
            LOG.debug("    certificate_chain: contents of %s", settings.ssl_settings.cert_file)

            ssl_credentials = ssl_channel_credentials(
                root_certificates=ca_cert, private_key=cert, certificate_chain=cert_chain
            )
        return secure_authorized_channel(
            credentials,
            RefreshRequester(),
            target,
            ssl_credentials=ssl_credentials,
            options=options,
        )

    if settings.ssl_settings:
        cert_chain = read_file_bytes(settings.ssl_settings.cert_file)
        cert = read_file_bytes(settings.ssl_settings.cert_key_file)
        ca_cert = read_file_bytes(settings.ssl_settings.ca_file)

        LOG.debug("Using a secure gRPC connection:")
        LOG.debug("    target: %s", target)
        LOG.debug("    root_certificates: contents of %s", settings.ssl_settings.ca_file)
        LOG.debug("    private_key: contents of %s", settings.ssl_settings.cert_key_file)
        LOG.debug("    certificate_chain: contents of %s", settings.ssl_settings.cert_file)

        credentials = ssl_channel_credentials(
            root_certificates=ca_cert, private_key=cert, certificate_chain=cert_chain
        )
        return secure_channel(target, credentials, options)
    else:
        LOG.debug("Using an insecure gRPC connection...")
        return insecure_channel(target, options)


class GRPCv1Connection(_LedgerConnection):
    def __init__(
        self,
        invoker: "Invoker",
        options: "LedgerConnectionOptions",
        settings: "HTTPConnectionSettings",
        context_path: "Optional[str]",
    ):
        super(GRPCv1Connection, self).__init__(invoker, options, settings, context_path)
        self._closed = Event()
        self._channel = grpc_create_channel(settings)
        self.active_contract_set_service = G_ActiveContractsServiceStub(self._channel)
        self.command_service = G_CommandServiceStub(self._channel)
        self.transaction_service = G_TransactionServiceStub(self._channel)
        self.package_service = G_PackageServiceStub(self._channel)
        self.package_management_service = G_PackageManagementServiceStub(self._channel)
        self.ledger_identity_service = G_LedgerIdentityServiceStub(self._channel)
        self.time_service = G_TimeServiceStub(self._channel)

    def close(self):
        try:
            LOG.info("Closing a GRPCv1Connection channel...")
            self._closed.set()
            self._channel.close()
        except:
            LOG.warning("An exception was thrown when trying to close down connections.")
        finally:
            super().close()

    @property
    def closed(self) -> bool:
        return self._closed.is_set()



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Type system types
------------------

The :mod:`dazl.model.types` module contains the Python classes used to represent the DAML type
system.

+----------------------+---------------------------+
| DAML type            | Python type               |
+======================+===========================+
| ``Bool``             | ``bool``                  |
+----------------------+---------------------------+
| ``Int``              | ``int``                   |
+----------------------+---------------------------+
| ``Decimal``          | ``decimal.Decimal``       |
+----------------------+---------------------------+
| ``[a]``              | ``list``                  |
+----------------------+---------------------------+

.. autoclass:: Type
.. autoclass:: ScalarType
.. autoclass:: ListType
.. autoclass:: RecordType
.. autoclass:: VariantType
.. autoclass:: UnsupportedType
"""
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Callable,
    ClassVar,
    Collection,
    Dict,
    Mapping,
    NewType,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)
import warnings

from .. import LOG
from ..damlast.daml_lf_1 import DottedName, Expr, ModuleRef, PackageRef, TypeConName
from ..prim import ContractData, Party
from ..util.typing import safe_cast, safe_dict_cast, safe_optional_cast

warnings.warn("The symbols in dazl.model.types are deprecated", DeprecationWarning, stacklevel=2)

DottedNameish = Union[str, Sequence[str]]

T_co = TypeVar("T_co", covariant=True)

if TYPE_CHECKING:
    from .types_store import PackageStore


# Reference to a ledger ID.
LedgerId = NewType("LedgerId", str)

# Reference to a package via a package identifier. The identifier is the ascii7
# lowercase hex-encoded hash of the package contents found in the DAML LF Archive.
#
# This is re-exported from the daml_lf_1 as PackageId for backwards compatibility, and will
# be removed in dazl v8.
PackageId = PackageRef

# A set of PackageId.
PackageIdSet = AbstractSet[PackageId]


def type_ref(s: str) -> "TypeReference":
    """
    Convenience method for creating a fully-qualified :class:`TypeReference`. A few formats are
    supported:

    "packageId:module:entity"
    "module:entity@pkgid"
    "module.entity@pkgid"
    """
    warnings.warn(
        "type_ref is deprecated; there is no replacement", DeprecationWarning, stacklevel=2
    )
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)

        # Attempt to parse in an older format still used by Navigator for display purposes.
        me, is_at, pkgid = s.partition("@")
        if is_at:
            # encourage people to use the new format
            m, has_colon, e = me.partition(":")
            if not has_colon:
                m, _, e = me.rpartition(".")

            pkg_ref = PackageRef(pkgid)
            module_name = DottedName(m.split("."))
            entity_name = e.split(".")
        else:
            components = s.split(":")
            if len(components) != 3:
                raise ValueError(f"could not parse as a template reference: {s!r}")

            pkg_ref = PackageRef(components[0])
            module_name = DottedName(components[1].split("."))
            entity_name = components[2].split(".")

        return TypeReference(
            con=TypeConName(module=ModuleRef(pkg_ref, module_name), name=entity_name)
        )


def dotted_name(obj: DottedNameish) -> "Sequence[str]":
    """
    Sanitize a string or a tuple of strings to a dotted name.

    :param obj: A string or tuple of strings.
    :return: A tuple of strings.
    """
    if obj is None:
        raise ValueError("DottedName must be a non-None value")
    if isinstance(obj, str):
        return tuple(obj.split("."))
    if isinstance(obj, Collection):
        for item in obj:
            if not isinstance(item, str):
                raise ValueError("DottedName's components must all be strings")
        return tuple(obj)
    else:
        raise ValueError("could not convert to a sequence of str: {obj!r}")


class NamedArgumentList(tuple):
    """
    A simple tuple to storing (name, value) pairs.
    """

    @property
    def names(self):
        return set(name for name, _ in self)


def type_dispatch_table(
    on_type_ref: Callable[["TypeReference"], T_co],
    on_type_var: Callable[["TypeVariable"], T_co],
    on_type_app: Callable[["TypeApp"], T_co],
    on_scalar: Callable[["ScalarType"], T_co],
    on_contract_id: Callable[["ContractIdType"], T_co],
    on_optional: Callable[["OptionalType"], T_co],
    on_list: Callable[["ListType"], T_co],
    on_text_map: "Callable[[TextMapType], T_co]",
    on_record: Callable[["RecordType"], T_co],
    on_variant: Callable[["VariantType"], T_co],
    on_enum: "Callable[[EnumType], T_co]",
    on_unsupported: Callable[["UnsupportedType"], T_co],
) -> Callable[["Type"], T_co]:
    def _impl(tt: Type):
        if isinstance(tt, TypeReference):
            return on_type_ref(tt)
        elif isinstance(tt, TypeVariable):
            return on_type_var(tt)
        elif isinstance(tt, TypeApp):
            return on_type_app(tt)
        elif isinstance(tt, ScalarType):
            return on_scalar(tt)
        elif isinstance(tt, ContractIdType):
            return on_contract_id(tt)
        elif isinstance(tt, OptionalType):
            return on_optional(tt)
        elif isinstance(tt, ListType):
            return on_list(tt)
        elif isinstance(tt, TextMapType):
            return on_text_map(tt)
        elif isinstance(tt, RecordType):
            return on_record(tt)
        elif isinstance(tt, VariantType):
            return on_variant(tt)
        elif isinstance(tt, EnumType):
            return on_enum(tt)
        elif isinstance(tt, UnsupportedType):
            return on_unsupported(tt)
        else:
            # note to maintainers: if you modify the Type hierarchy, you must also maintain this
            # poor man's pattern match over the hierarchy
            LOG.error("Incomplete implementation of type_match! (when handling %r)", tt)
            raise Exception(f"unknown Type subclass: {tt!r}")

    return _impl


def scalar_type_dispatch_table(
    on_unit: "Callable[[], T_co]",
    on_bool: "Callable[[], T_co]",
    on_text: "Callable[[], T_co]",
    on_int: "Callable[[], T_co]",
    on_decimal: "Callable[[], T_co]",
    on_party: "Callable[[], T_co]",
    on_date: "Callable[[], T_co]",
    on_datetime: "Callable[[], T_co]",
    on_timedelta: "Callable[[], T_co]",
) -> "Callable[[ScalarType], T_co]":
    def _impl(tt: ScalarType):
        st = safe_cast(ScalarType, tt)
        if st == SCALAR_TYPE_UNIT:
            return on_unit()
        elif tt == SCALAR_TYPE_BOOL:
            return on_bool()
        elif tt == SCALAR_TYPE_TEXT or tt == SCALAR_TYPE_CHAR:
            return on_text()
        elif tt == SCALAR_TYPE_INTEGER:
            return on_int()
        elif tt == SCALAR_TYPE_DECIMAL or tt == SCALAR_TYPE_NUMERIC:
            return on_decimal()
        elif tt == SCALAR_TYPE_PARTY:
            return on_party()
        elif tt == SCALAR_TYPE_DATE:
            return on_date()
        elif tt == SCALAR_TYPE_DATETIME:
            return on_datetime()
        elif tt == SCALAR_TYPE_RELTIME:
            return on_timedelta()
        else:
            # note to maintainers: if you modify the set of ScalarType instances, you must also
            # maintain this poor man's pattern match over the hierarchy
            LOG.error(
                "Incomplete implementation of scalar_type_dispatch_table! (when handling %r)", tt
            )
            raise Exception(f"unknown ScalarType: {tt!r}")

    return _impl


class Type:
    """
    A DAML-defined type.
    """

    def __init__(self):
        warnings.warn(
            "dazl.model.types.Type and its subclasses are deprecated. "
            "Use dazl.damlast.daml_lf_1.Type instead.",
            DeprecationWarning,
            stacklevel=2,
        )

    def __str__(self):
        from ..pretty import DAML_PRETTY_PRINTER

        return DAML_PRETTY_PRINTER.visit_type(self)


class TypeApp(Type):
    """
    An application of one or more types to an open type.

    This is basically Type.Con in the Protobuf declaration.
    """

    def __init__(self, body: Type, arguments: Sequence[Type]):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            # noinspection PyDeprecation
            super().__init__()

        warnings.warn(
            "dazl.model.types.TypeApp and its subclasses are deprecated. "
            "Use dazl.damlast.daml_lf_1.Type instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        if not isinstance(body, Type):
            raise ValueError(f"a Type is required here (got {body} instead)")
        if not arguments:
            raise ValueError("at least one type argument is required in a TypeApp")
        self.body = body
        self.arguments = tuple(arguments)

    def __repr__(self):
        return f"<TypeApp(body={self.body}, arguments={self.arguments}>"


class TypeVariable(Type):
    """
    An unbound type in a Type expression.
    """

    __slots__ = ("name",)

    def __init__(self, name: str):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            # noinspection PyDeprecation
            super().__init__()

        warnings.warn(
            "dazl.model.types.TypeVar is deprecated. " "Use dazl.damlast.daml_lf_1.Type instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.name = safe_cast(str, name)

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"TypeVariable({self.name})"

    def __eq__(self, other):
        return isinstance(other, TypeVariable) and self.name == other.name

    def __hash__(self):
        return hash(self.name)


class TypeReference(Type):
    def __init__(self, con: "TypeConName"):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            # noinspection PyDeprecation
            super().__init__()
        warnings.warn(
            "TypeReference is deprecated; in most cases, use TypeConName instead",
            DeprecationWarning,
            stacklevel=2,
        )
        self.con = con

    def __str__(self):
        return str(self.con)

    def __eq__(self, other):
        return isinstance(other, TypeReference) and self.con == other.con

    def __hash__(self):
        return hash(self.con)


class UnresolvedTypeReference(Type):
    """
    A reference that may or may not ultimately resolve to a type.
    """

    def __init__(self, name: str):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            # noinspection PyDeprecation
            super().__init__()
        warnings.warn(
            "UnresolvedTypeReference is deprecated; in most cases, use TypeConName instead",
            DeprecationWarning,
            stacklevel=2,
        )
        self.name = safe_cast(str, name)

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"<UnresolvedTypeReference({self.name!r})>"

    def __eq__(self, other):
        return isinstance(other, UnresolvedTypeReference) and self.name == other.name

    def __hash__(self):
        return hash((UnresolvedTypeReference, self.name))


class ConcreteType(Type):
    pass


class ScalarType(ConcreteType):
    """
    A DAML-defined type that represents a simple scalar value. You should not need to ever
    construct instances of this directly; all scalar types are builtins.
    """

    __slots__ = ("name",)
    name: str
    BUILTINS: "ClassVar[Sequence[ScalarType]]"

    def __init__(self, name: str):
        """
        Construct an object that references a scalar DAML type.

        :param name: The name of this type as it is known in DAML.
        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            # noinspection PyDeprecation
            super().__init__()
        warnings.warn(
            "ScalarType is deprecated; in most cases, use dazl.damlast.daml_lf_1 types instead",
            DeprecationWarning,
            stacklevel=2,
        )
        self.name = safe_cast(str, name)

    def __repr__(self):
        """
        Return the DAML name of this type.
        """
        return self.name

    def __hash__(self):
        return hash((ScalarType, self.name))

    def __eq__(self, other):
        return isinstance(other, ScalarType) and self.name == other.name


class _BuiltInParameterizedType(ConcreteType):
    """
    Convenience class that encapsulates commonalities for the built-in types that have one type
    parameter.
    """

    __slots__ = ("type_parameter",)

    def __init__(self, type_parameter: Type):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            # noinspection PyDeprecation
            super().__init__()

        warnings.warn(
            "ContractIdType, and ListType, TypeRefType are deprecated; use dazl.damlast.daml_lf_1 types instead",
            DeprecationWarning,
            stacklevel=2,
        )
        self.type_parameter = safe_cast(Type, type_parameter)

    def __repr__(self):
        py_type = type(self).__name__
        return f"<{py_type}({self.type_parameter!r})>"


class ContractIdType(_BuiltInParameterizedType):
    pass


class ListType(_BuiltInParameterizedType):
    pass


class TypeRefType(_BuiltInParameterizedType):
    pass


class TextMapType(ConcreteType):
    """
    A DAML-defined TextMap.

    Instance attributes:

    .. attribute: TextMapType.value_type

        The type of values in this map.
    """

    __slots__ = ("value_type",)

    def __init__(self, value_type: Type):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            # noinspection PyDeprecation
            super().__init__()
        warnings.warn(
            "TextMapType is deprecated; use dazl.damlast.daml_lf_1 types instead",
            DeprecationWarning,
            stacklevel=2,
        )

        self.value_type = safe_cast(Type, value_type)

    def __repr__(self):
        py_type = type(self).__name__
        return f"<{py_type}({self.value_type!r})>"


class GenMapType(ConcreteType):
    """
    A DAML-defined GenMap.

    Instance attributes:

    .. attribute: TextMapType.key_type

        The type of keys in this map.

    .. attribute: TextMapType.value_type

        The type of values in this map.
    """

    __slots__ = "key_type", "value_type"

    def __init__(self, key_type: Type, value_type: Type):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            # noinspection PyDeprecation
            super().__init__()
        warnings.warn(
            "GenMapType is deprecated; use dazl.damlast.daml_lf_1 types instead",
            DeprecationWarning,
            stacklevel=2,
        )

        self.key_type = safe_cast(Type, key_type)
        self.value_type = safe_cast(Type, value_type)

    def __repr__(self):
        py_type = type(self).__name__
        return f"<{py_type}({self.key_type!r}, {self.value_type!r})>"


class OptionalType(_BuiltInParameterizedType):
    """
    A DAML-defined Optional.

    Instance attributes:

    .. attribute: OptionalType.type_parameter

        The type of value in the Optional.
    """


class UpdateType(_BuiltInParameterizedType):
    pass


class ForAllType(Type):
    def __init__(self, type_vars, body_type):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            # noinspection PyDeprecation
            super().__init__()
        warnings.warn(
            "ForAllType is deprecated; there is no replacement", DeprecationWarning, stacklevel=2
        )

        self.type_vars = type_vars
        self.body_type = body_type

    def __repr__(self):
        return f"<ForAllType({self.type_vars}, {self.body_type})>"


class _CompositeDataType(ConcreteType):
    """
    Either a :class:`RecordType` (product type) or a :class:`VariantType` (sum type).
    """

    def __init__(
        self,
        named_args: "NamedArgumentList",
        name: "Optional[TypeReference]",
        type_args: "Sequence[TypeVariable]",
    ):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            # noinspection PyDeprecation
            super().__init__()
        warnings.warn(
            "RecordType and VariantType are deprecated; use dazl.damlast.daml_lf_1 types instead",
            DeprecationWarning,
            stacklevel=2,
        )

        if type(self) == _CompositeDataType:
            raise Exception("_CompositeDataType cannot be constructed")
        if not isinstance(named_args, NamedArgumentList):
            raise TypeError("NamedArgumentList required here")
        if name is not None and not isinstance(name, TypeReference):
            raise TypeError("name must be a TypeReference or None")

        self.named_args = named_args
        self.name = name
        self.type_args = tuple(type_args)

    def field_type(self, name: str) -> Type:
        for key, value_type in self.named_args:
            if key == name:
                return value_type

        raise ValueError(f"field or constructor {name!r} not found in {self}")

    def __repr__(self):
        py_type = type(self).__name__

        name = "(anonymous)" if self.name is None else self.name
        full_name = "".join(f" {v}" for v in ((name,) + self.type_args))

        return f"<{py_type}:{full_name} {self.named_args}>"


class FunctionType(Type):
    """
    Representation of a DAML function signature.

    Instances of this type aren't practically usable from this library. They are merely recorded in
    order to faithfully pretty-print metadata.
    """

    def __init__(self, parameters: Sequence[Type], result: Type):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            # noinspection PyDeprecation
            super().__init__()
        warnings.warn(
            "FunctionType is deprecated; there is no replacement", DeprecationWarning, stacklevel=2
        )

        self.parameters = tuple(parameters)
        self.result = safe_cast(Type, result)

    def __str__(self):
        from io import StringIO

        with StringIO() as buf:
            for param in self.parameters:
                buf.write(str(param))
                buf.write(" -> ")
            buf.write(str(self.result))
            return buf.getvalue()

    def __repr__(self):
        return f"<FunctionType({self})>"


class RecordType(_CompositeDataType):
    def as_args_list(self):
        return self.named_args


class VariantType(_CompositeDataType):
    def as_args_list(self):
        return self.named_args

    def _find_ctor(self, constructor_name: str) -> Type:
        return self.field_type(constructor_name)


class EnumType(ConcreteType):

    __slots__ = ("constructors",)

    def __init__(self, name: "Optional[TypeReference]", constructors: "Collection[str]"):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            # noinspection PyDeprecation
            super().__init__()

        warnings.warn(
            "EnumType is deprecated; use dazl.damlast.daml_lf_1 types instead",
            DeprecationWarning,
            stacklevel=2,
        )

        self.name = name
        self.constructors = constructors


class UnsupportedType(Type):
    """
    A DAML type that is currently unparseable by the Python client library.
    """

    __slots__ = ("name",)

    def __init__(self, name):
        self.name = safe_cast(str, name)

    def __repr__(self):
        return f"<UnsupportedType({self.name})>"


class Template:
    """
    Definition of a contract template.
    """

    def __init__(
        self,
        data_type: "RecordType",
        key_type: "Optional[Type]",
        choices: "Collection[TemplateChoice]",
        observers: "Expr",
        signatories: "Expr",
        agreement: "Expr",
        ensure: "Expr",
    ):
        if not isinstance(data_type, RecordType) or data_type.name is None:
            raise ValueError(
                f"data_type is required and must be a named record type " f"(got {data_type})"
            )
        self.data_type = safe_cast(RecordType, data_type)
        self.key_type = safe_optional_cast(Type, key_type)
        self.choices = choices
        self._observers = observers
        self._signatories = signatories
        self._agreement = agreement
        self._ensure = ensure


class TemplateChoice:
    __slots__ = ("name", "consuming", "data_type", "return_type", "_controllers")

    def __init__(
        self, name: str, consuming: bool, data_type: Type, return_type: Type, controllers: "Expr"
    ):
        self.name = name
        self.consuming = consuming
        self.data_type = data_type
        self.return_type = return_type
        self._controllers = controllers

    @property
    def type(self):
        return self.data_type

    def controllers(self, cdata: ContractData) -> Collection[Party]:
        """
        Return every :class:`Party` that can exercise this choice given the specified contract data.
        """


def as_commands(commands_ish, allow_callables=False):
    """
    Converts something that is either ``None``, a single :class:`Command`, or an iterable over
    :class:`Command` objects to a ``list`` of :class:`Command`.

    :param commands_ish:
        Something that might be construed as either a :class:`Command` or an iterable over
        :class:`Command`.
    :param allow_callables:
        If callables are encountered, invoke them and expect them to return something that can be
        easily serialized to a :class:`Command`.
    """
    from .writing import Command

    if commands_ish is None:
        return ()
    elif isinstance(commands_ish, Command):
        return (commands_ish,)
    elif allow_callables and callable(commands_ish):
        return as_commands(commands_ish(), allow_callables=False)

    # assume this is some kind of iterable structure, where everything needs to be a Command
    cmds = []
    for command in commands_ish:
        if allow_callables and callable(command):
            cmds.extend(as_commands(command(), allow_callables=False))
        elif isinstance(command, Command):
            cmds.append(command)
        else:
            raise TypeError(f"{command!r} is not a Command")
    return tuple(cmds)


def as_contract_id(cid, template_id=None):
    """
    Convert something that resembles a contract ID to a :class:`ContractId` or
    :class:`RelativeContractRef`.
    """
    warnings.warn(
        "as_contract_id is deprecated; there is no replacement", DeprecationWarning, stacklevel=2
    )
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        from .core import ContractId

        if cid is None:
            raise ValueError("cid is required")
        elif isinstance(cid, str):
            return ContractId(cid, template_id)
        elif isinstance(cid, ContractId):
            return cid

        raise TypeError("Could not serialize an object to a contract ID: {!r}".format(cid))


with warnings.catch_warnings():
    warnings.simplefilter("ignore", DeprecationWarning)
    SCALAR_TYPE_UNIT = ScalarType("Unit")
    SCALAR_TYPE_BOOL = ScalarType("Bool")
    SCALAR_TYPE_CHAR = ScalarType("Char")
    SCALAR_TYPE_INTEGER = ScalarType("Integer")
    SCALAR_TYPE_DECIMAL = ScalarType("Decimal")
    SCALAR_TYPE_NUMERIC = ScalarType("Numeric")
    SCALAR_TYPE_TEXT = ScalarType("Text")
    SCALAR_TYPE_PARTY = ScalarType("Party")
    SCALAR_TYPE_RELTIME = ScalarType("RelTime")
    SCALAR_TYPE_DATE = ScalarType("Date")
    SCALAR_TYPE_TIME = ScalarType("Time")
    SCALAR_TYPE_ANY = ScalarType("Any")
    SCALAR_TYPE_DATETIME = SCALAR_TYPE_TIME

ScalarType.BUILTINS = [
    SCALAR_TYPE_BOOL,
    SCALAR_TYPE_CHAR,
    SCALAR_TYPE_INTEGER,
    SCALAR_TYPE_DECIMAL,
    SCALAR_TYPE_NUMERIC,
    SCALAR_TYPE_TEXT,
    SCALAR_TYPE_PARTY,
    SCALAR_TYPE_RELTIME,
    SCALAR_TYPE_DATE,
    SCALAR_TYPE_TIME,
    SCALAR_TYPE_ANY,
]


class TypeEvaluationContext:
    references: Dict[TypeConName, ConcreteType]
    variables: Dict[TypeVariable, Type]
    path: Sequence[Union[TypeReference, str]]

    __slots__ = ("references", "variables", "path")

    @classmethod
    def from_store(cls, store: "PackageStore") -> "TypeEvaluationContext":
        return cls(store.types(), {}, ())

    def __init__(self, references: "Mapping[TypeConName, ConcreteType]", variables, path):
        self.references = safe_dict_cast(TypeConName, ConcreteType, references)
        self.variables = safe_dict_cast(TypeVariable, Type, variables)
        self.path = path

    def append_path(self, component: Union[TypeReference, str]) -> "TypeEvaluationContext":
        return TypeEvaluationContext(
            references=self.references,
            variables=self.variables,
            path=tuple((*self.path, component)),
        )

    def resolve_var(self, var: TypeVariable) -> Type:
        return self.variables[var]

    def with_vars(self, new_vars: "Dict[TypeVariable, Type]") -> "TypeEvaluationContext":
        confirmed_new_vars = {}
        for new_var, new_var_value in new_vars.items():
            if new_var == new_var_value:
                # TODO: Why do these cases happen?
                continue
            confirmed_new_vars[new_var] = new_var_value

        return TypeEvaluationContext(
            references=self.references,
            variables={**self.variables, **confirmed_new_vars},
            path=self.path,
        )


def type_evaluate_dispatch(
    on_scalar: "Callable[[TypeEvaluationContext, ScalarType], T_co]",
    on_contract_id: "Callable[[TypeEvaluationContext,  ContractIdType], T_co]",
    on_optional: "Callable[[TypeEvaluationContext, OptionalType], T_co]",
    on_list: "Callable[[TypeEvaluationContext, ListType], T_co]",
    on_text_map: "Callable[[TypeEvaluationContext, TextMapType], T_co]",
    on_record: "Callable[[TypeEvaluationContext, RecordType], T_co]",
    on_variant: "Callable[[TypeEvaluationContext, VariantType], T_co]",
    on_enum: "Callable[[TypeEvaluationContext, EnumType], T_co]",
    on_unsupported: "Callable[[TypeEvaluationContext, UnsupportedType], T_co]",
) -> "Callable[[TypeEvaluationContext, Type], T_co]":
    """
    Produce a function that defers handling of core types to the passed in functions.

    The cases of :class:`TypeReference, :class:`TypeApp`, and :class:`TypeVariable` are handled
    automatically. Note, though that ultimately type evaluation is only performed at one level
    deep, and the produced function may need to be called multiple types at multiple depths of an
    object or type hierarchy.
    """

    def _impl(context, tt):
        resolve_depth = 0
        while isinstance(tt, (TypeReference, TypeVariable, TypeApp)):
            context, tt = single_reduce(context, tt)
            resolve_depth += 1
            if resolve_depth > 10:
                raise Exception("hit our max resolve depth, which is probably not so great")

        def error(_: Any) -> "T_co":
            raise Exception()

        context, tt = annotate_context(context, tt)

        return type_dispatch_table(
            error,
            error,
            error,
            lambda st: on_scalar(context, st),
            lambda ct: on_contract_id(context, ct),
            lambda ot: on_optional(context, ot),
            lambda lt: on_list(context, lt),
            lambda mt: on_text_map(context, mt),
            lambda rt: on_record(context, rt),
            lambda vt: on_variant(context, vt),
            lambda et: on_enum(context, et),
            lambda ut: on_unsupported(context, tt),
        )(tt)

    return _impl


def _type_evaluate_dispatch_error(_, __):
    raise Exception()


def type_evaluate_dispatch_default_error(
    on_scalar: "Callable[[TypeEvaluationContext, ScalarType], T_co]" = _type_evaluate_dispatch_error,
    on_contract_id: "Callable[[TypeEvaluationContext,  ContractIdType], T_co]" = _type_evaluate_dispatch_error,
    on_optional: "Callable[[TypeEvaluationContext,  OptionalType], T_co]" = _type_evaluate_dispatch_error,
    on_list: "Callable[[TypeEvaluationContext, ListType], T_co]" = _type_evaluate_dispatch_error,
    on_text_map: "Callable[[TypeEvaluationContext, TextMapType], T_co]" = _type_evaluate_dispatch_error,
    on_record: "Callable[[TypeEvaluationContext, RecordType], T_co]" = _type_evaluate_dispatch_error,
    on_variant: "Callable[[TypeEvaluationContext, VariantType], T_co]" = _type_evaluate_dispatch_error,
    on_enum: "Callable[[TypeEvaluationContext, EnumType], T_co]" = _type_evaluate_dispatch_error,
    on_unsupported: "Callable[[TypeEvaluationContext, UnsupportedType], T_co]" = _type_evaluate_dispatch_error,
):
    return type_evaluate_dispatch(
        on_scalar,
        on_contract_id,
        on_optional,
        on_list,
        on_text_map,
        on_record,
        on_variant,
        on_enum,
        on_unsupported,
    )


def single_reduce(context: TypeEvaluationContext, tt: Type) -> "Tuple[TypeEvaluationContext, Type]":
    """
    Apply a single substitution/reduction/unwrapping. The context may be augmented with additional
    variables if a TypeApp is encountered.
    """

    def identity(t):
        return context, t

    def reduce_app(ta: TypeApp) -> "Tuple[TypeEvaluationContext, Type]":
        body = context.references[ta.body.con] if isinstance(ta.body, TypeReference) else ta.body
        if not isinstance(body, (RecordType, VariantType)):
            raise Exception("Can't apply types to non-generic data structures")

        return context.with_vars(dict(zip(body.type_args, ta.arguments))), ta.body

    return type_dispatch_table(
        lambda tr: (context, context.references[tr.con]),
        lambda tv: (context, context.resolve_var(tv)),
        reduce_app,
        identity,
        identity,
        identity,
        identity,
        identity,
        identity,
        identity,
        identity,
        identity,
    )(tt)


def annotate_context(
    context: TypeEvaluationContext, tt: Type
) -> Tuple[TypeEvaluationContext, Type]:
    def identity(t):
        return context, t

    def error(_: Any) -> Any:
        raise Exception()

    def annotate_path(t: Union[RecordType, VariantType]) -> Tuple[TypeEvaluationContext, Type]:
        if t.name is None:
            raise ValueError("cannot annotate a path with an anonymous type")

        return context.append_path(t.name), t

    return type_dispatch_table(
        error,
        error,
        error,
        identity,
        identity,
        identity,
        identity,
        identity,
        annotate_path,
        annotate_path,
        identity,
        identity,
    )(tt)


# types that can be used to refer to templates
TemplateNameLike = Union[str, TypeConName, TypeReference, UnresolvedTypeReference, Template]



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off

from typing import (
    AsyncIterable as __AsyncStream,
    Awaitable as __Awaitable,
    Iterable as __Stream,
    Union as __Union,
    overload,
)

from google.protobuf.empty_pb2 import Empty
from grpc import Channel as __Channel
from grpc.aio import Channel as __AsyncChannel

from .command_service_pb2 import (
    SubmitAndWaitForTransactionIdResponse,
    SubmitAndWaitForTransactionResponse,
    SubmitAndWaitForTransactionTreeResponse,
    SubmitAndWaitRequest,
)

__all__ = [
    "CommandServiceStub",
]

class CommandServiceStub:
    @classmethod
    @overload
    def __new__(cls, channel: __Channel) -> _CommandSubmissionServiceStub: ...
    @classmethod
    @overload
    def __new__(cls, channel: __AsyncChannel) -> _CommandSubmissionServiceStub_Async: ...
    def SubmitAndWait(self, __1: SubmitAndWaitRequest) -> __Union[Empty, __Awaitable[Empty]]: ...
    def SubmitAndWaitForTransactionId(self, __1: SubmitAndWaitRequest) -> __Union[SubmitAndWaitForTransactionIdResponse, __Awaitable[SubmitAndWaitForTransactionIdResponse]]: ...
    def SubmitAndWaitForTransaction(self, __1: SubmitAndWaitRequest) -> __Union[SubmitAndWaitForTransactionResponse, __Awaitable[SubmitAndWaitForTransactionResponse]]: ...
    def SubmitAndWaitForTransactionTree(self, __1: SubmitAndWaitRequest) -> __Union[SubmitAndWaitForTransactionTreeResponse, __Awaitable[SubmitAndWaitForTransactionTreeResponse]]: ...

class _CommandSubmissionServiceStub(CommandServiceStub):
    def SubmitAndWait(self, __1: SubmitAndWaitRequest) -> Empty: ...
    def SubmitAndWaitForTransactionId(self, __1: SubmitAndWaitRequest) -> SubmitAndWaitForTransactionIdResponse: ...
    def SubmitAndWaitForTransaction(self, __1: SubmitAndWaitRequest) -> SubmitAndWaitForTransactionResponse: ...
    def SubmitAndWaitForTransactionTree(self, __1: SubmitAndWaitRequest) -> SubmitAndWaitForTransactionTreeResponse: ...

class _CommandSubmissionServiceStub_Async(CommandServiceStub):
    def SubmitAndWait(self, __1: SubmitAndWaitRequest) -> __Awaitable[Empty]: ...
    def SubmitAndWaitForTransactionId(self, __1: SubmitAndWaitRequest) -> __Awaitable[SubmitAndWaitForTransactionIdResponse]: ...
    def SubmitAndWaitForTransaction(self, __1: SubmitAndWaitRequest) -> __Awaitable[SubmitAndWaitForTransactionResponse]: ...
    def SubmitAndWaitForTransactionTree(self, __1: SubmitAndWaitRequest) -> __Awaitable[SubmitAndWaitForTransactionTreeResponse]: ...



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off

from typing import (
    Any as __Any,
    Dict as __Map,
    List as __List,
    Literal as __Literal,
    Mapping as __Mapping,
    NoReturn as __Void,
    Optional as __Optional,
    Sequence as __Sequence,
    Tuple as __Tuple,
    overload,
)

from google.protobuf.descriptor import FieldDescriptor as __FieldDescriptor
from google.protobuf.message import Message as __Message

from .......prim import Party
from .ledger_offset_pb2 import LedgerOffset
from .trace_context_pb2 import TraceContext
from .transaction_filter_pb2 import TransactionFilter
from .transaction_pb2 import Transaction, TransactionTree

__all__ = [
    "GetTransactionsRequest",
    "GetTransactionsResponse",
    "GetTransactionTreesResponse",
    "GetTransactionByEventIdRequest",
    "GetTransactionByIdRequest",
    "GetTransactionResponse",
    "GetFlatTransactionResponse",
    "GetLedgerEndRequest",
    "GetLedgerEndResponse",
]

class GetTransactionsRequest(__Message):
    ledger_id: str
    begin: LedgerOffset
    end: LedgerOffset
    filter: TransactionFilter
    verbose: bool
    trace_context: TraceContext
    def __init__(self, *, ledger_id: __Optional[str] = ..., begin: __Optional[LedgerOffset] = ..., end: __Optional[LedgerOffset] = ..., filter: __Optional[TransactionFilter] = ..., verbose: __Optional[bool] = ..., trace_context: __Optional[TraceContext] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: GetTransactionsRequest) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["ledger_id", "begin", "end", "filter", "verbose", "trace_context"]) -> bool: ...
    def ClearField(self, field_name: __Literal["ledger_id", "begin", "end", "filter", "verbose", "trace_context"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class GetTransactionsResponse(__Message):
    transactions: __List[Transaction]
    def __init__(self, *, transactions: __Optional[__Sequence[Transaction]] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: GetTransactionResponse) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["transactions"]) -> bool: ...
    def ClearField(self, field_name: __Literal["transactions"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class GetTransactionTreesResponse(__Message):
    transactions: __List[TransactionTree]
    def __init__(self, *, transactions: __Optional[__Sequence[TransactionTree]] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: GetTransactionTreesResponse) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["transactions"]) -> bool: ...
    def ClearField(self, field_name: __Literal["transactions"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class GetTransactionByEventIdRequest(__Message):
    ledger_id: str
    event_id: str
    requesting_parties: __List[Party]
    trace_context: TraceContext
    def __init__(self, *, ledger_id: __Optional[str] = ..., event_id: __Optional[str] = ..., requesting_parties: __Optional[__Sequence[Party]] = ..., trace_context: __Optional[TraceContext] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: GetTransactionByEventIdRequest) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["ledger_id", "event_id", "requesting_parties", "trace_context"]) -> bool: ...
    def ClearField(self, field_name: __Literal["ledger_id", "event_id", "requesting_parties", "trace_context"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class GetTransactionByIdRequest(__Message):
    ledger_id: str
    event_id: str
    requesting_parties: __List[Party]
    trace_context: TraceContext
    def __init__(self, *, ledger_id: __Optional[str] = ..., event_id: __Optional[str] = ..., requesting_parties: __Optional[__Sequence[Party]] = ..., trace_context: __Optional[TraceContext] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: GetTransactionByIdRequest) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["ledger_id", "event_id", "requesting_parties", "trace_context"]) -> bool: ...
    def ClearField(self, field_name: __Literal["ledger_id", "event_id", "requesting_parties", "trace_context"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class GetTransactionResponse(__Message):
    transaction: TransactionTree
    def __init__(self, *, transaction: __Optional[TransactionTree] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: GetTransactionResponse) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["transaction"]) -> bool: ...
    def ClearField(self, field_name: __Literal["transaction"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class GetFlatTransactionResponse(__Message):
    transaction: Transaction
    def __init__(self, *, transaction: __Optional[Transaction] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: GetFlatTransactionResponse) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["transaction"]) -> bool: ...
    def ClearField(self, field_name: __Literal["transaction"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class GetLedgerEndRequest(__Message):
    ledger_id: str
    trace_context: TraceContext
    def __init__(self, *, ledger_id: __Optional[str] = ..., trace_context: __Optional[TraceContext] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: GetLedgerEndRequest) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["ledger_id", "trace_context"]) -> bool: ...
    def ClearField(self, field_name: __Literal["ledger_id", "trace_context"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class GetLedgerEndResponse(__Message):
    offset: LedgerOffset
    def __init__(self, *, offset: __Optional[LedgerOffset] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: GetLedgerEndResponse) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["offset"]) -> bool: ...
    def ClearField(self, field_name: __Literal["offset"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from collections import defaultdict
from dataclasses import dataclass
from functools import reduce
from operator import add
import sys
from threading import RLock
from types import MappingProxyType
from typing import (
    Any,
    Collection,
    DefaultDict,
    Dict,
    Generic,
    Iterable,
    List,
    Mapping,
    Optional,
    TypeVar,
    Union,
    no_type_check,
)
import warnings

from ..damlast.daml_lf_1 import Archive, Expr, Package, PackageRef, TypeConName, ValName, _Name
from ..damlast.lookup import validate_template
from ..util.typing import safe_cast, safe_dict_cast

if sys.version_info >= (3, 8):
    from typing import Protocol, runtime_checkable
else:
    from typing_extensions import Protocol, runtime_checkable


with warnings.catch_warnings():
    warnings.simplefilter("ignore", DeprecationWarning)

    from .types import (
        ConcreteType,
        PackageIdSet,
        Template,
        TemplateChoice,
        Type,
        TypeReference,
        UnresolvedTypeReference,
    )

K = TypeVar("K", bound=_Name)
T = TypeVar("T")
S = TypeVar("S")

warnings.warn(
    "The types of dazl.model.types_store are deprecated", DeprecationWarning, stacklevel=2
)


class PackageStoreBuilder:
    """
    Convenience class for building up a :class:`PackageStore`.
    """

    def __init__(self):
        warnings.warn(
            "PackageStoreBuilder is deprecated; there is no direct replacement.",
            DeprecationWarning,
            stacklevel=2,
        )

        self._archives = list()  # type: List[Archive]
        self._value_types = dict()  # type: Dict[ValName, Expr]
        self._data_types = dict()  # type: Dict[TypeConName, Type]
        self._templates = dict()  # type: Dict[TypeConName, Template]
        self._expected_package_ids = None  # type: Optional[Collection[PackageRef]]

    def add_archive(self, archive: "Archive") -> None:
        self._archives.append(archive)

    def add_type(self, name: "TypeConName", data_type: Type):
        safe_cast(TypeConName, name)
        self._data_types[name] = safe_cast(Type, data_type)

    def add_value(self, name: "ValName", value: "Expr"):
        safe_cast(ValName, name)
        self._value_types[name] = safe_cast(Expr, value)

    def add_template(self, template: Template):
        name = template.data_type.name
        if name is None:
            raise ValueError("template cannot be defined with an anonymous type")
        self._templates[name.con] = safe_cast(Template, template)

    def add_expected_package_ids(self, expected_package_ids: "Collection[PackageRef]"):
        self._expected_package_ids = expected_package_ids

    def get_type(self, name: TypeConName) -> "Optional[Type]":
        return self._data_types.get(name)

    def build(self) -> "PackageStore":
        return PackageStore(
            self._archives,
            self._value_types,
            self._data_types,
            self._templates,
            self._expected_package_ids,
        )


class PackageStore:
    """
    A thread-safe store of type information.
    """

    @classmethod
    def empty(cls):
        """
        Create an empty store.
        """
        warnings.warn(
            "PackageStore is deprecated; use PackageLoader/SymbolLookup instead",
            DeprecationWarning,
            stacklevel=2,
        )

        return cls([], {}, {}, {}, None)

    def __init__(
        self,
        archives: "Collection[Archive]",
        value_types: "Dict[ValName, Expr]",
        data_types: "Dict[TypeConName, Type]",
        templates: "Dict[TypeConName, Template]",
        expected_package_ids: "Optional[Collection[PackageRef]]" = None,
    ):
        warnings.warn(
            "PackageStore is deprecated; use PackageLoader/SymbolLookup instead",
            DeprecationWarning,
            stacklevel=2,
        )
        self._lock = RLock()
        self._archives = list(archives)
        self._cache = PackageStoreCache(EMPTY_TYPE_CACHE, EMPTY_TYPE_CACHE, EMPTY_TYPE_CACHE)
        self._value_types = safe_dict_cast(ValName, Expr, value_types)
        self._data_types = safe_dict_cast(TypeConName, Type, data_types)
        self._templates = safe_dict_cast(TypeConName, Template, templates)
        self._expected_package_ids = expected_package_ids

    def archives(self) -> "Collection[Archive]":
        """
        Return a copy of the collection of the set of loaded :class:`Archive`s.
        """
        warnings.warn(
            "PackageStore.archives() is deprecated; use SymbolLookup.archives() instead",
            DeprecationWarning,
            stacklevel=2,
        )

        with self._lock:
            return list(self._archives)

    def packages(self) -> "Collection[Package]":
        """
        Return a copy of the collection of the set of loaded :class:`Package`s.
        """
        warnings.warn(
            "PackageStore.packages() is deprecated; use SymbolLookup.archives() instead",
            DeprecationWarning,
            stacklevel=2,
        )

        with self._lock:
            return [a.package for a in self._archives]

    def package_ids(self) -> "Collection[PackageRef]":
        """
        Return a copy of the collection of the set of loaded :class:`Package`s.
        """
        warnings.warn(
            "PackageStore.package_ids() is deprecated; use SymbolLookup.package_ids() instead",
            DeprecationWarning,
            stacklevel=2,
        )

        with self._lock:
            return [a.hash for a in self._archives]

    def expected_package_ids(self) -> "Optional[Collection[PackageRef]]":
        """
        Return package IDs that are expected to be found on the ledger.
        """
        warnings.warn(
            "PackageStore.expected_package_ids() is deprecated; there is no replacement.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._expected_package_ids

    def register_all(self, other_store: "PackageStore") -> "PackageStore":
        """
        Register all types declared in the other :class:`PackageStore`.

        :param other_store: A package store to copy types, templates, and choices from.
        :return: A reference to this object.
        """
        warnings.warn(
            "PackageStore.register_all() is deprecated; use PackageLoader instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        if self is not other_store:
            with self._lock:
                self._archives.extend(other_store._archives)
                self._value_types.update(other_store._value_types)
                self._data_types.update(other_store._data_types)
                self._templates.update(other_store._templates)
                self._cache = PackageStoreCache(
                    TypeCache.build(self._value_types),
                    TypeCache.build(self._data_types),
                    TypeCache.build(self._templates),
                )
        return self

    def resolve_value_reference(self, value_ref: "ValName") -> "Expr":
        warnings.warn(
            "PackageStore.resolve_value_reference() is deprecated; "
            "use SymbolLookup.value() instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        with self._lock:
            return self._value_types[value_ref]

    def resolve_type_reference(self, template_ref: "Union[TypeConName, TypeReference]") -> Type:
        """
        Resolve a type based on its reference.

        :param template_ref:
            A :class:`TypeReference` that refers to a type.
        :return:
            The :class:`Type` that is referred to by this type reference.
        :raise KeyError:
            If the :class:`TypeReference` does not have a corresponding value in this store.
        """
        warnings.warn(
            "PackageStore.resolve_value_reference() is deprecated; "
            "use SymbolLookup.data_type() instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        if isinstance(template_ref, TypeReference):
            con = template_ref.con
        elif isinstance(template_ref, TypeConName):
            con = template_ref
        else:
            raise ValueError("template_ref must either by a TypeConName or TypeReference")

        with self._lock:
            return self._data_types[con]

    def types(self) -> "Mapping[TypeConName, ConcreteType]":
        warnings.warn(
            'PackageStore.types() is deprecated; use SymbolLookup.data_type("*") instead.',
            DeprecationWarning,
            stacklevel=2,
        )

        with self._lock:
            return MappingProxyType(dict(self._data_types))  # type: ignore

    def get_templates_for_packages(
        self, package_ids: "Iterable[PackageRef]"
    ) -> "Collection[Template]":
        """
        Return a collection of :class:`Template` instances from the given set of package IDs.

        :param package_ids:
            The set of package IDs to restrict templates to.
        :return:
            A collection of matching templates, or an empty collection if none match. This method
            never returns ``None``.
        """
        warnings.warn(
            "PackageStore.get_templates_for_packages() is deprecated; "
            'use [SymbolLookup.templates(f"{pkg}:*") for pkg in package_ids] instead.',
            DeprecationWarning,
            stacklevel=2,
        )

        match = []  # type: List[Template]
        for pkg_id in package_ids:
            match.extend(self._cache.templates.lookup(pkg_id, "*"))
        return match

    def resolve_template(
        self, template: "Union[None, str, TypeReference, UnresolvedTypeReference, TypeConName]"
    ) -> Collection[Template]:
        """
        Return a collection of :class:`Template` instances that match the specified template name.

        Some special parameters:
         * If ``"*"`` is passed in, all templates are returned.
         * If ``None`` is passed in, an empty collection is returned.

        :param template:
            A template name or a :class:`TypeReference`.
        :return:
            A collection of matching templates, or an empty collection if none match. This method
            never returns ``None``.
        """
        warnings.warn(
            "PackageStore.resolve_template() is deprecated; use SymbolLookup.template() instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        if isinstance(template, Template):
            # if we were given a Template for some strange reason, just simply return a single-item
            # tuple of that given Template
            return [template]

        package_id, template_name = validate_template(template, allow_deprecated_identifiers=True)
        return self._cache.templates.lookup(package_id, template_name)

    def resolve_template_type(
        self, template: "Union[None, str, TypeReference, UnresolvedTypeReference, TypeConName]"
    ) -> Dict[TypeReference, Type]:
        """
        Return a collection of types that match for the template.

        :param template:
            A template name or a :class:`TypeReference`.
        :return:
            A dictionary of possible matches, or empty if there are no matches. This method never
            returns ``None``.
        """
        warnings.warn(
            "PackageStore.resolve_template() is deprecated; use SymbolLookup.template() instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)

            return {
                template.data_type.name: template.data_type
                for template in self.resolve_template(template)
                if template.data_type.name is not None
            }

    def resolve_choice(self, template: Any, choice: str) -> Dict[TypeReference, TemplateChoice]:
        """
        Return all possible choices for the combination of template identifier and choice name.
        If template is ``'*'`` or ``None``, all choices with the specified name are returned.
        """
        warnings.warn(
            'PackageStore.resolve_choice() is deprecated; use SymbolLookup.template("*") instead, '
            "and lookup the choice within the returned Template.",
            DeprecationWarning,
            stacklevel=2,
        )

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)

            matches = dict()  # type: Dict[TypeReference, TemplateChoice]
            for t in self.resolve_template(template):
                for c in t.choices:
                    if c.name == choice and t.data_type.name is not None:
                        matches[t.data_type.name] = c
            return matches


@runtime_checkable
class PackageProvider(Protocol):
    """
    Interface to an object that can provide package information.
    """

    def __init__(self):
        warnings.warn(
            "PackageProvider is deprecated; use PackageLookup instead.",
            DeprecationWarning,
            stacklevel=2,
        )

    def get_package_ids(self) -> "PackageIdSet":
        """
        Return the current universe of package IDs.
        """
        raise NotImplementedError

    def fetch_package(self, package_id: "PackageRef") -> bytes:
        """
        Retrieve the bytes that correspond to a package.
        """
        raise NotImplementedError

    def get_all_packages(self) -> "Mapping[PackageRef, bytes]":
        raise NotImplementedError


class MemoryPackageProvider:
    def __init__(self, mapping: "Mapping[PackageRef, bytes]"):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            super().__init__()
        warnings.warn(
            "MemoryPackageProvider is deprecated; there is no replacement.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.mapping = mapping

    def get_package_ids(self) -> "PackageIdSet":
        return frozenset(self.mapping.keys())

    # fetch_package is actually supposed to throw an error on missing packages, but people may
    # be reliant on the old behavior; anyway this entire file is deprecated; changing the behavior
    # of a soon-to-be-removed class seems unnecessary
    @no_type_check
    def fetch_package(self, package_id: "PackageRef") -> bytes:
        return self.mapping.get(package_id)

    def get_all_packages(self) -> "Mapping[PackageRef, bytes]":
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            return {pkg_id: self.fetch_package(pkg_id) for pkg_id in self.get_package_ids()}


@dataclass(frozen=True)
class PackageStoreCache:
    value_types: "TypeCache[ValName, Expr]"
    data_types: "TypeCache[TypeConName, Type]"
    templates: "TypeCache[TypeConName, Template]"


@dataclass(frozen=True)
class TypeCache(Generic[K, T]):
    """
    Immutable cache of type information.

    Instance attributes:

    .. attribute:: TypeCache.by_package_lookup

        A mapping from package ID strings to another mapping, where keys are fully-qualified type
        names for templates to a collection of matching Template instances. These collections are
        never non-empty; they will only ever contain multiple entries if there is a collision in
        dot-separated names.

    .. attribute:: TypeCache.by_name_lookup

        A mapping from fully-qualified type names (excluding package IDs) to ``Collection[T]``.
        These subcollections will only contain multiple values if names are not unique with respect
        to package IDs.
    """

    everything: "Collection[T]"
    by_package_lookup: "Mapping[PackageRef, Mapping[str, Collection[T]]]"
    by_name_lookup: "Mapping[str, Collection[T]]"

    @classmethod
    def build(cls, objects: "Mapping[K, T]") -> "TypeCache[K, T]":
        from ..damlast.util import package_local_name, package_ref

        everything = tuple(objects.values())
        by_package_lookup: DefaultDict[PackageRef, DefaultDict[str, List[T]]] = defaultdict(
            lambda: defaultdict(list)
        )
        by_name_lookup = defaultdict(list)

        for k, v in objects.items():
            package_id = package_ref(k)
            module_entity_name = package_local_name(k)
            # This is kept here for backwards compatibility, but its use should be discouraged
            module_entity_name_deprecated = module_entity_name.replace(":", ".")
            for valid_name in (module_entity_name, module_entity_name_deprecated):
                by_package_lookup[package_id][valid_name].append(v)
                by_name_lookup[valid_name].append(v)

        return TypeCache(
            everything,
            _immutable_mmc(by_package_lookup),
            MappingProxyType({k: tuple(v) for k, v in by_name_lookup.items()}),
        )

    def lookup(self, package_id: PackageRef, type_name: str) -> "Collection[T]":
        """
        Look up items based on package ID and type name. The values cannot be ``None``, but they
        can be the special-meaning ``'*'`` value.
        """
        safe_cast(str, package_id)
        safe_cast(str, type_name)

        if package_id == "*":
            if type_name == "*":
                # *:* means return everything
                return self.everything
            else:
                # *:Specific.Module:Entity means return all matches of the requested name, but
                # under any package
                return self.by_name_lookup.get(type_name, ())
        elif type_name == "*":
            # PKG_ID:* means return all matches in a specific package
            return reduce(add, self.by_package_lookup.get(package_id, {}).values(), ())
        else:
            # PKG_ID:Specific.Module:Entity; we are looking for a very specific type
            return self.by_package_lookup.get(package_id, {}).get(type_name, ())


def _immutable_mmc(
    mapping: "Mapping[S, Mapping[str, Collection[T]]]",
) -> "Mapping[S, Mapping[str, Collection[T]]]":
    """
    Create an immutable copy of :class:`TemplateStoreCache` data structures.
    """
    return MappingProxyType(
        {k1: MappingProxyType({k2: tuple(v) for k2, v in v1.items()}) for k1, v1 in mapping.items()}
    )


EMPTY_MAPPING = MappingProxyType({})  # type: ignore
EMPTY_TYPE_CACHE = TypeCache((), EMPTY_MAPPING, EMPTY_MAPPING)  # type: ignore



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Mapping, Tuple

from ..damlast.daml_lf_1 import DefDataType, Type
from ..prim import (
    to_bool,
    to_date,
    to_datetime,
    to_decimal,
    to_int,
    to_party,
    to_record,
    to_str,
    to_variant,
)
from .context import Context
from .mapper import ValueMapper

__all__ = ["CanonicalMapper"]


class CanonicalMapper(ValueMapper):
    """
    A mapper that canonicalizes values. If values are already in canonical form, this is essentially
    the identity mapper.

    For container primitives (Optional, List, Map, etc), records, and variants, values are recursed
    and walked through, so subclasses that override :class:`CanonicalMapper` can typically just
    override the primitive methods in order to get slightly different behavior.

    The canonical format of DAML-LF values in dazl was designed specifically to conform as closely
    as possible to simple JSON representations, so the :class:`CanonicalMapper` can also be used to
    decode DAML-LF JSON.
    """

    def data_record(
        self, context: "Context", dt: "DefDataType", record: "DefDataType.Fields", obj: "Any"
    ) -> "Any":
        orig_mapping = self._record_to_dict(context, dt, record, obj)

        expected_keys = frozenset(fld.field for fld in record.fields)
        actual_keys = frozenset(orig_mapping)
        if actual_keys.issuperset(expected_keys):
            if actual_keys != expected_keys:
                # Earlier versions of dazl were more tolerant of extra fields. To keep backwards
                # compatibility, we'll emit a warning, though this may become an exception
                # eventually.
                context.value_warn(
                    obj, f'extra fields: {", ".join(sorted(actual_keys - expected_keys))}'
                )
        else:
            context.value_error(
                obj, f'missing fields: {", ".join(sorted(expected_keys - actual_keys))}'
            )

        new_mapping = {
            fld.field: context.append_path(fld.field).convert(fld.type, orig_mapping[fld.field])
            for fld in record.fields
        }
        return self._dict_to_record(context, dt, record, new_mapping)

    def data_variant(
        self, context: "Context", dt: "DefDataType", variant: "DefDataType.Fields", obj: "Any"
    ) -> "Any":
        ctor, orig_val = self._variant_to_ctor_value(context, dt, variant, obj)
        for fld in variant.fields:
            if fld.field == ctor:
                new_val = context.append_path(fld.field).convert(fld.type, orig_val)
                return self._ctor_value_to_variant(context, dt, variant, ctor, new_val)

        # searched through all fields, and did not find the constructor
        raise ValueError(f"could not find a variant constructor for {ctor}")

    def data_enum(
        self,
        context: "Context",
        dt: "DefDataType",
        enum: "DefDataType.EnumConstructors",
        obj: "Any",
    ) -> "Any":
        return context.value_validate_enum(obj, enum)

    def prim_unit(self, context: "Context", obj: "Any") -> "Any":
        return {}

    def prim_bool(self, context: "Context", obj: "Any") -> "Any":
        return to_bool(obj)

    def prim_int64(self, context: "Context", obj: "Any") -> "Any":
        return to_int(obj)

    def prim_text(self, context: "Context", obj: "Any") -> "Any":
        return to_str(obj)

    def prim_timestamp(self, context: "Context", obj: "Any") -> "Any":
        return to_datetime(obj)

    def prim_party(self, context: "Context", obj: "Any") -> "Any":
        return to_party(obj)

    def prim_list(self, context: "Context", item_type: "Type", obj: "Any") -> "Any":
        return context.convert_list(item_type, obj)

    def prim_date(self, context: "Context", obj: "Any") -> "Any":
        return to_date(obj)

    def prim_contract_id(self, context: "Context", item_type: "Type", obj: "Any") -> "Any":
        return context.convert_contract_id(item_type, obj)

    def prim_optional(self, context: "Context", item_type: "Type", obj: "Any") -> "Any":
        return context.convert_optional(item_type, obj)

    def prim_text_map(self, context: "Context", item_type: "Type", obj: "Any") -> "Any":
        return context.convert_text_map(item_type, obj)

    def prim_numeric(self, context: "Context", nat: int, obj: "Any") -> "Any":
        return to_decimal(obj)

    def prim_gen_map(
        self, context: "Context", key_type: "Type", value_type: "Type", obj: "Any"
    ) -> "Any":
        return obj

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def _record_to_dict(
        self, context: "Context", dt: "DefDataType", record: "DefDataType.Fields", obj: "Any"
    ) -> "Mapping[str, Any]":
        """
        Convert a record object to a Python dict. Should be overridden by subclasses to convert a
        record to a dict whose keys are field names and values are associated field values if record
        objects are not always understood to be dicts.

        The default implementation assumes that ``obj`` is already a ``dict`` that matches this
        contract and simply returns it (though this is verified first).
        """
        return to_record(obj)

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def _dict_to_record(
        self, context: "Context", dt: "DefDataType", record: "DefDataType.Fields", obj: "Any"
    ):
        return obj

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def _variant_to_ctor_value(
        self, context: "Context", dt: "DefDataType", record: "DefDataType.Fields", obj: "Any"
    ) -> "Tuple[str, Any]":
        """
        Convert a variant object to a constructor and a value. Should be overridden by subclasses to
        convert a variant that is not formatted in a recognized way.
        """
        return to_variant(obj)

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def _ctor_value_to_variant(
        self,
        context: "Context",
        dt: "DefDataType",
        variant: "DefDataType.Fields",
        ctor: str,
        value: "Any",
    ) -> "Any":
        return {ctor: value}



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/admin/party_management_service.proto

from google.protobuf import (
    descriptor as _descriptor,
    message as _message,
    reflection as _reflection,
    symbol_database as _symbol_database,
)

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="com/daml/ledger/api/v1/admin/party_management_service.proto",
    package="com.daml.ledger.api.v1.admin",
    syntax="proto3",
    serialized_options=b"\n\034com.daml.ledger.api.v1.adminB PartyManagementServiceOuterClass\252\002\034Com.Daml.Ledger.Api.V1.Admin",
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n;com/daml/ledger/api/v1/admin/party_management_service.proto\x12\x1c\x63om.daml.ledger.api.v1.admin"\x19\n\x17GetParticipantIdRequest"2\n\x18GetParticipantIdResponse\x12\x16\n\x0eparticipant_id\x18\x01 \x01(\t"$\n\x11GetPartiesRequest\x12\x0f\n\x07parties\x18\x01 \x03(\t"W\n\x12GetPartiesResponse\x12\x41\n\rparty_details\x18\x01 \x03(\x0b\x32*.com.daml.ledger.api.v1.admin.PartyDetails"\x19\n\x17ListKnownPartiesRequest"]\n\x18ListKnownPartiesResponse\x12\x41\n\rparty_details\x18\x01 \x03(\x0b\x32*.com.daml.ledger.api.v1.admin.PartyDetails"C\n\x14\x41llocatePartyRequest\x12\x15\n\rparty_id_hint\x18\x01 \x01(\t\x12\x14\n\x0c\x64isplay_name\x18\x02 \x01(\t"Z\n\x15\x41llocatePartyResponse\x12\x41\n\rparty_details\x18\x01 \x01(\x0b\x32*.com.daml.ledger.api.v1.admin.PartyDetails"E\n\x0cPartyDetails\x12\r\n\x05party\x18\x01 \x01(\t\x12\x14\n\x0c\x64isplay_name\x18\x02 \x01(\t\x12\x10\n\x08is_local\x18\x03 \x01(\x08\x32\x8b\x04\n\x16PartyManagementService\x12\x81\x01\n\x10GetParticipantId\x12\x35.com.daml.ledger.api.v1.admin.GetParticipantIdRequest\x1a\x36.com.daml.ledger.api.v1.admin.GetParticipantIdResponse\x12o\n\nGetParties\x12/.com.daml.ledger.api.v1.admin.GetPartiesRequest\x1a\x30.com.daml.ledger.api.v1.admin.GetPartiesResponse\x12\x81\x01\n\x10ListKnownParties\x12\x35.com.daml.ledger.api.v1.admin.ListKnownPartiesRequest\x1a\x36.com.daml.ledger.api.v1.admin.ListKnownPartiesResponse\x12x\n\rAllocateParty\x12\x32.com.daml.ledger.api.v1.admin.AllocatePartyRequest\x1a\x33.com.daml.ledger.api.v1.admin.AllocatePartyResponseB_\n\x1c\x63om.daml.ledger.api.v1.adminB PartyManagementServiceOuterClass\xaa\x02\x1c\x43om.Daml.Ledger.Api.V1.Adminb\x06proto3',
)


_GETPARTICIPANTIDREQUEST = _descriptor.Descriptor(
    name="GetParticipantIdRequest",
    full_name="com.daml.ledger.api.v1.admin.GetParticipantIdRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=93,
    serialized_end=118,
)


_GETPARTICIPANTIDRESPONSE = _descriptor.Descriptor(
    name="GetParticipantIdResponse",
    full_name="com.daml.ledger.api.v1.admin.GetParticipantIdResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="participant_id",
            full_name="com.daml.ledger.api.v1.admin.GetParticipantIdResponse.participant_id",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=120,
    serialized_end=170,
)


_GETPARTIESREQUEST = _descriptor.Descriptor(
    name="GetPartiesRequest",
    full_name="com.daml.ledger.api.v1.admin.GetPartiesRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="parties",
            full_name="com.daml.ledger.api.v1.admin.GetPartiesRequest.parties",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=172,
    serialized_end=208,
)


_GETPARTIESRESPONSE = _descriptor.Descriptor(
    name="GetPartiesResponse",
    full_name="com.daml.ledger.api.v1.admin.GetPartiesResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="party_details",
            full_name="com.daml.ledger.api.v1.admin.GetPartiesResponse.party_details",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=210,
    serialized_end=297,
)


_LISTKNOWNPARTIESREQUEST = _descriptor.Descriptor(
    name="ListKnownPartiesRequest",
    full_name="com.daml.ledger.api.v1.admin.ListKnownPartiesRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=299,
    serialized_end=324,
)


_LISTKNOWNPARTIESRESPONSE = _descriptor.Descriptor(
    name="ListKnownPartiesResponse",
    full_name="com.daml.ledger.api.v1.admin.ListKnownPartiesResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="party_details",
            full_name="com.daml.ledger.api.v1.admin.ListKnownPartiesResponse.party_details",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=326,
    serialized_end=419,
)


_ALLOCATEPARTYREQUEST = _descriptor.Descriptor(
    name="AllocatePartyRequest",
    full_name="com.daml.ledger.api.v1.admin.AllocatePartyRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="party_id_hint",
            full_name="com.daml.ledger.api.v1.admin.AllocatePartyRequest.party_id_hint",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="display_name",
            full_name="com.daml.ledger.api.v1.admin.AllocatePartyRequest.display_name",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=421,
    serialized_end=488,
)


_ALLOCATEPARTYRESPONSE = _descriptor.Descriptor(
    name="AllocatePartyResponse",
    full_name="com.daml.ledger.api.v1.admin.AllocatePartyResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="party_details",
            full_name="com.daml.ledger.api.v1.admin.AllocatePartyResponse.party_details",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=490,
    serialized_end=580,
)


_PARTYDETAILS = _descriptor.Descriptor(
    name="PartyDetails",
    full_name="com.daml.ledger.api.v1.admin.PartyDetails",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="party",
            full_name="com.daml.ledger.api.v1.admin.PartyDetails.party",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="display_name",
            full_name="com.daml.ledger.api.v1.admin.PartyDetails.display_name",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="is_local",
            full_name="com.daml.ledger.api.v1.admin.PartyDetails.is_local",
            index=2,
            number=3,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=582,
    serialized_end=651,
)

_GETPARTIESRESPONSE.fields_by_name["party_details"].message_type = _PARTYDETAILS
_LISTKNOWNPARTIESRESPONSE.fields_by_name["party_details"].message_type = _PARTYDETAILS
_ALLOCATEPARTYRESPONSE.fields_by_name["party_details"].message_type = _PARTYDETAILS
DESCRIPTOR.message_types_by_name["GetParticipantIdRequest"] = _GETPARTICIPANTIDREQUEST
DESCRIPTOR.message_types_by_name["GetParticipantIdResponse"] = _GETPARTICIPANTIDRESPONSE
DESCRIPTOR.message_types_by_name["GetPartiesRequest"] = _GETPARTIESREQUEST
DESCRIPTOR.message_types_by_name["GetPartiesResponse"] = _GETPARTIESRESPONSE
DESCRIPTOR.message_types_by_name["ListKnownPartiesRequest"] = _LISTKNOWNPARTIESREQUEST
DESCRIPTOR.message_types_by_name["ListKnownPartiesResponse"] = _LISTKNOWNPARTIESRESPONSE
DESCRIPTOR.message_types_by_name["AllocatePartyRequest"] = _ALLOCATEPARTYREQUEST
DESCRIPTOR.message_types_by_name["AllocatePartyResponse"] = _ALLOCATEPARTYRESPONSE
DESCRIPTOR.message_types_by_name["PartyDetails"] = _PARTYDETAILS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

GetParticipantIdRequest = _reflection.GeneratedProtocolMessageType(
    "GetParticipantIdRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETPARTICIPANTIDREQUEST,
        "__module__": "com.daml.ledger.api.v1.admin.party_management_service_pb2"
        # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.GetParticipantIdRequest)
    },
)
_sym_db.RegisterMessage(GetParticipantIdRequest)

GetParticipantIdResponse = _reflection.GeneratedProtocolMessageType(
    "GetParticipantIdResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETPARTICIPANTIDRESPONSE,
        "__module__": "com.daml.ledger.api.v1.admin.party_management_service_pb2"
        # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.GetParticipantIdResponse)
    },
)
_sym_db.RegisterMessage(GetParticipantIdResponse)

GetPartiesRequest = _reflection.GeneratedProtocolMessageType(
    "GetPartiesRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETPARTIESREQUEST,
        "__module__": "com.daml.ledger.api.v1.admin.party_management_service_pb2"
        # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.GetPartiesRequest)
    },
)
_sym_db.RegisterMessage(GetPartiesRequest)

GetPartiesResponse = _reflection.GeneratedProtocolMessageType(
    "GetPartiesResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETPARTIESRESPONSE,
        "__module__": "com.daml.ledger.api.v1.admin.party_management_service_pb2"
        # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.GetPartiesResponse)
    },
)
_sym_db.RegisterMessage(GetPartiesResponse)

ListKnownPartiesRequest = _reflection.GeneratedProtocolMessageType(
    "ListKnownPartiesRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _LISTKNOWNPARTIESREQUEST,
        "__module__": "com.daml.ledger.api.v1.admin.party_management_service_pb2"
        # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.ListKnownPartiesRequest)
    },
)
_sym_db.RegisterMessage(ListKnownPartiesRequest)

ListKnownPartiesResponse = _reflection.GeneratedProtocolMessageType(
    "ListKnownPartiesResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _LISTKNOWNPARTIESRESPONSE,
        "__module__": "com.daml.ledger.api.v1.admin.party_management_service_pb2"
        # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.ListKnownPartiesResponse)
    },
)
_sym_db.RegisterMessage(ListKnownPartiesResponse)

AllocatePartyRequest = _reflection.GeneratedProtocolMessageType(
    "AllocatePartyRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _ALLOCATEPARTYREQUEST,
        "__module__": "com.daml.ledger.api.v1.admin.party_management_service_pb2"
        # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.AllocatePartyRequest)
    },
)
_sym_db.RegisterMessage(AllocatePartyRequest)

AllocatePartyResponse = _reflection.GeneratedProtocolMessageType(
    "AllocatePartyResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _ALLOCATEPARTYRESPONSE,
        "__module__": "com.daml.ledger.api.v1.admin.party_management_service_pb2"
        # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.AllocatePartyResponse)
    },
)
_sym_db.RegisterMessage(AllocatePartyResponse)

PartyDetails = _reflection.GeneratedProtocolMessageType(
    "PartyDetails",
    (_message.Message,),
    {
        "DESCRIPTOR": _PARTYDETAILS,
        "__module__": "com.daml.ledger.api.v1.admin.party_management_service_pb2"
        # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.PartyDetails)
    },
)
_sym_db.RegisterMessage(PartyDetails)


DESCRIPTOR._options = None

_PARTYMANAGEMENTSERVICE = _descriptor.ServiceDescriptor(
    name="PartyManagementService",
    full_name="com.daml.ledger.api.v1.admin.PartyManagementService",
    file=DESCRIPTOR,
    index=0,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_start=654,
    serialized_end=1177,
    methods=[
        _descriptor.MethodDescriptor(
            name="GetParticipantId",
            full_name="com.daml.ledger.api.v1.admin.PartyManagementService.GetParticipantId",
            index=0,
            containing_service=None,
            input_type=_GETPARTICIPANTIDREQUEST,
            output_type=_GETPARTICIPANTIDRESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="GetParties",
            full_name="com.daml.ledger.api.v1.admin.PartyManagementService.GetParties",
            index=1,
            containing_service=None,
            input_type=_GETPARTIESREQUEST,
            output_type=_GETPARTIESRESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="ListKnownParties",
            full_name="com.daml.ledger.api.v1.admin.PartyManagementService.ListKnownParties",
            index=2,
            containing_service=None,
            input_type=_LISTKNOWNPARTIESREQUEST,
            output_type=_LISTKNOWNPARTIESRESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="AllocateParty",
            full_name="com.daml.ledger.api.v1.admin.PartyManagementService.AllocateParty",
            index=3,
            containing_service=None,
            input_type=_ALLOCATEPARTYREQUEST,
            output_type=_ALLOCATEPARTYRESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
)
_sym_db.RegisterServiceDescriptor(_PARTYMANAGEMENTSERVICE)

DESCRIPTOR.services_by_name["PartyManagementService"] = _PARTYMANAGEMENTSERVICE

# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
Command types that are used in the dazl v5 API.

These symbols are primarily kept for backwards compatibility.
"""
from dataclasses import dataclass, fields
from datetime import timedelta
from typing import Any, Collection, List, Mapping, Optional, Sequence, Tuple, Union
import uuid
import warnings

from ..damlast.daml_lf_1 import TypeConName
from ..ledger import api_types
from ..prim import ContractData, ContractId, Party

__all__ = [
    "CommandBuilder",
    "CommandDefaults",
    "CommandPayload",
    "CommandsOrCommandSequence",
    "CreateAndExerciseCommand",
    "CreateCommand",
    "EventHandlerResponse",
    "ExerciseByKeyCommand",
    "ExerciseCommand",
    "create",
    "create_and_exercise",
    "exercise",
    "exercise_by_key",
    "flatten_command_sequence",
]


class CreateCommand(api_types.CreateCommand):
    """
    A command that creates a contract without any predecessors.
    """

    def __init__(
        self, template: "Union[str, TypeConName]", arguments: "Optional[ContractData]" = None
    ):
        warnings.warn(
            "dazl.client.commands.CreateCommand is deprecated; "
            "prefer calling dazl.ledger.Connection.create or "
            "dazl.client.PartyClient.submit_create, "
            "or use dazl.ledger.CreateCommand instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(template, arguments or {})

    @property
    def template_type(self) -> "TypeConName":
        """
        Use :meth:`template_id` instead.
        """
        warnings.warn(
            "CreateCommand.template_type is deprecated; use CreateCommand.template_id instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.template_id

    @property
    def arguments(self) -> "Mapping[str, Any]":
        """
        Use :meth:`payload` instead.
        """
        warnings.warn(
            "CreateCommand.arguments is deprecated; use CreateCommand.payload instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.payload


class CreateAndExerciseCommand(api_types.CreateAndExerciseCommand):
    """
    A command that exercises a choice on a newly-created contract, in a single transaction.
    """

    def __init__(
        self,
        template: "Union[str, TypeConName]",
        arguments: "Mapping[str, Any]",
        choice: str,
        choice_argument: "Optional[Any]" = None,
    ):
        warnings.warn(
            "dazl.client.commands.CreateAndExerciseCommand is deprecated; "
            "prefer calling dazl.ledger.Connection.create_and_exercise or "
            "dazl.client.PartyClient.submit_create_and_exercise, "
            "or use dazl.ledger.CreateAndExerciseCommand instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(template, arguments, choice, choice_argument)

    @property
    def template_type(self) -> "TypeConName":
        """
        Use :meth:`template_id` instead.
        """
        warnings.warn(
            "CreateAndExerciseCommand.template_type is deprecated; "
            "use CreateAndExerciseCommand.template_id instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.template_id

    @property
    def arguments(self) -> "Any":
        """
        Use :meth:`payload` instead.
        """
        warnings.warn(
            "CreateAndExerciseCommand.arguments is deprecated; "
            "use CreateAndExerciseCommand.payload instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.payload

    @property
    def choice_argument(self) -> "Any":
        """
        Use :meth:`argument` instead.
        """
        warnings.warn(
            "CreateAndExerciseCommand.choice_argument is deprecated; "
            "use CreateAndExerciseCommand.argument instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.argument


class ExerciseCommand(api_types.ExerciseCommand):
    """
    A command that exercises a choice on a contract identified by its contract ID.
    """

    def __init__(self, contract: "ContractId", choice: str, arguments: "Optional[Any]" = None):
        warnings.warn(
            "dazl.client.commands.ExerciseCommand is deprecated; "
            "prefer calling dazl.ledger.Connection.exercise or "
            "dazl.client.PartyClient.submit_exercise, "
            "or use dazl.ledger.ExerciseCommand instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(contract, choice, arguments)

    @property
    def contract(self) -> "ContractId":
        """
        Use :meth:`contract_id` instead.
        """
        warnings.warn(
            "ExerciseCommand.contract is deprecated; use ExerciseCommand.contract_id instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.contract_id

    @property
    def arguments(self) -> "Any":
        """
        Use :meth:`argument` instead.
        """
        warnings.warn(
            "ExerciseCommand.arguments is deprecated; use ExerciseCommand.argument instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.argument


class ExerciseByKeyCommand(api_types.ExerciseByKeyCommand):
    def __init__(
        self,
        template: "Union[str, TypeConName]",
        contract_key: "Any",
        choice: str,
        choice_argument: "Optional[Any]" = None,
    ):
        warnings.warn(
            "dazl.client.commands.ExerciseByKeyCommand is deprecated; "
            "prefer calling dazl.ledger.Connection.exercise_by_key or "
            "dazl.client.PartyClient.submit_exercise_by_key, "
            "or use dazl.ledger.ExerciseByKeyCommand instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(template, contract_key, choice, choice_argument)

    @property
    def template_type(self) -> "TypeConName":
        """
        Use :meth:`template_id` instead.
        """
        warnings.warn(
            "ExerciseByKeyCommand.template_type is deprecated; "
            "use ExerciseByKeyCommand.template_id instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.template_id

    @property
    def contract_key(self) -> "Any":
        """
        Use :meth:`argument` instead.
        """
        warnings.warn(
            "ExerciseByKeyCommand.contract_key is deprecated; "
            "use ExerciseByKeyCommand.key instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        return self.key

    @property
    def choice_argument(self) -> "Any":
        """
        Use :meth:`argument` instead.
        """
        warnings.warn(
            "ExerciseByKeyCommand.choice_argument is deprecated; "
            "use ExerciseByKeyCommand.argument instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.argument


CommandsOrCommandSequence = Union[None, api_types.Command, Sequence[Optional[api_types.Command]]]


# noinspection PyDeprecation
class CommandBuilder:
    """
    Builder class for generating commands to be sent to the ledger.
    """

    @classmethod
    def coerce(cls, obj, atomic_default=False) -> "CommandBuilder":
        """
        Create a :class:`CommandBuilder` from the objects that an event handler is allowed to
        return.

        :param obj:
        :param atomic_default:
        :return:
        """
        warnings.warn(
            "CommandBuilder is deprecated; "
            "prefer calling dazl.ledger.Connection.commands, "
            "dazl.client.PartyClient.submit, or construct commands explicitly instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            if isinstance(obj, CommandBuilder):
                return obj

            builder = CommandBuilder(atomic_default=atomic_default)
            if obj is not None:
                builder.append(obj)
            return builder

    def __init__(self, atomic_default: bool = False):
        warnings.warn(
            "CommandBuilder is deprecated; "
            "prefer calling dazl.ledger.Connection.commands, "
            "dazl.client.PartyClient.submit, or construct commands explicitly instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self._atomic_default = atomic_default
        self._commands = [[]]  # type: List[List[api_types.Command]]
        self._defaults = CommandDefaults()

    def defaults(
        self,
        party: "Optional[Party]" = None,
        ledger_id: "Optional[str]" = None,
        workflow_id: "Optional[str]" = None,
        application_id: "Optional[str]" = None,
        command_id: "Optional[str]" = None,
        deduplication_time: "Optional[timedelta]" = None,
    ) -> None:
        if party is not None:
            self._defaults.default_party = party
        if ledger_id is not None:
            self._defaults.default_ledger_id = ledger_id
        if workflow_id is not None:
            self._defaults.default_workflow_id = workflow_id
        if application_id is not None:
            self._defaults.default_application_id = application_id
        if command_id is not None:
            self._defaults.default_command_id = command_id
        if deduplication_time is not None:
            self._defaults.default_deduplication_time = deduplication_time

    def create(self, template, arguments=None) -> "CommandBuilder":
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            return self.append(create(template, arguments=arguments))

    def exercise(self, contract, choice, arguments=None) -> "CommandBuilder":
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            return self.append(exercise(contract, choice, arguments=arguments))

    def create_and_exercise(
        self, template, create_arguments, choice_name, choice_arguments=None
    ) -> "CommandBuilder":
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            return self.append(
                create_and_exercise(template, create_arguments, choice_name, choice_arguments)
            )

    def append(self, *commands: "CommandsOrCommandSequence") -> "CommandBuilder":
        """
        Append one or more commands, or list of commands to the :class:`CommandBuilder` in flight.
        This method respects the value of ``atomic_default`` that this object was constructed with.
        In order to force commands to be submitted either atomically, use :meth:`append_atomically`.
        To allow these commands to be submitted in parallel use :meth:`append_nonatomically`.

        :param commands: One or more commands, or list of commands to be submitted to the ledger.
        :return: This object.
        """
        if self._atomic_default:
            # a command builder that defaults to being atomic will put all commands in a single
            # transaction; build on the very first transaction
            self._commands[0].extend(flatten_command_sequence(commands))
            return self
        else:
            return self.append_nonatomically(*commands)

    def append_atomically(self, *commands: "CommandsOrCommandSequence") -> "CommandBuilder":
        self._commands.extend([flatten_command_sequence(commands)])
        return self

    def append_nonatomically(self, *commands: "CommandsOrCommandSequence") -> "CommandBuilder":
        self._commands.extend([[cmd] for cmd in flatten_command_sequence(commands)])
        return self

    def build(self, defaults: "Optional[CommandDefaults]" = None) -> "Collection[CommandPayload]":
        """
        Return a collection of commands.
        """
        if defaults is None:
            raise ValueError("defaults must currently be specified")

        command_id = (
            defaults.default_command_id or self._defaults.default_command_id or uuid.uuid4().hex
        )

        cmds = []  # type: List[CommandPayload]
        for i, commands in enumerate(self._commands):
            if commands:
                party = defaults.default_party or self._defaults.default_party
                ledger_id = defaults.default_ledger_id or self._defaults.default_ledger_id
                workflow_id = defaults.default_workflow_id or self._defaults.default_workflow_id
                application_id = (
                    defaults.default_application_id or self._defaults.default_application_id
                )
                deduplication_time = (
                    defaults.default_deduplication_time or self._defaults.default_deduplication_time
                )

                # hitting any of these checks would be quite unusual as this information is
                # validated and supplied in places higher up in the code; they're primarily here to
                # make sure this function typechecks
                if not party:
                    raise ValueError("party is missing on a Command")
                if not ledger_id:
                    raise ValueError("ledger_id is missing on a Command")
                if not workflow_id:
                    raise ValueError("workflow_id is missing on a Command")
                if not application_id:
                    raise ValueError("application_id is missing on a Command")

                cmds.append(
                    CommandPayload(
                        party=party,
                        ledger_id=ledger_id,
                        workflow_id=workflow_id,
                        application_id=application_id,
                        command_id=command_id,
                        deduplication_time=deduplication_time,
                        commands=commands,
                    )
                )

        return cmds

    def __format__(self, format_spec):
        if format_spec == "c":
            return str(self._commands)
        else:
            return repr(self)

    def __repr__(self):
        return f"CommandBuilder({self._commands})"


def flatten_command_sequence(
    commands: "Sequence[CommandsOrCommandSequence]",
) -> "List[api_types.Command]":
    """
    Convert a list of mixed commands, ``None``, and list of commands into an ordered sequence of
    non-``None`` commands.
    """
    ret = []  # type: List[api_types.Command]
    errors = []  # type: List[Tuple[Sequence[int], CommandsOrCommandSequence]]

    for i, obj in enumerate(commands):
        if obj is not None:
            if isinstance(obj, api_types.Command):
                ret.append(obj)
            else:
                try:
                    cmd_iter = iter(obj)
                except TypeError:
                    errors.append(((i,), obj))
                    continue
                for j, cmd in enumerate(cmd_iter):
                    if isinstance(cmd, api_types.Command):
                        ret.append(cmd)
                    else:
                        errors.append(((i, j), cmd))
    if errors:
        raise ValueError(
            f"Failed to interpret some elements as Commands in the list: " f"$[{index}] = {command}"
            for index, command in errors
        )
    return ret


@dataclass
class CommandDefaults:
    """
    Values to use for a :class:`Command` when no value is specified with the creation of the
    command.
    """

    default_party: Optional[Party] = None
    default_ledger_id: Optional[str] = None
    default_workflow_id: Optional[str] = None
    default_application_id: Optional[str] = None
    default_command_id: Optional[str] = None
    default_deduplication_time: Optional[timedelta] = None


@dataclass(frozen=True)
class CommandPayload:
    """
    A request to mutate active state of the ledger.

    .. attribute:: CommandPayload.party
        The party submitting the request.
    .. attribute:: CommandPayload.application_id:
        An optional application ID to accompany the request.
    .. attribute:: CommandPayload.command_id:
        A hash that represents the BIM commitment.
    .. attribute:: CommandPayload.deduplication_time:
        The maximum time interval before the client should consider this command expired.
    .. attribute:: CommandPayload.commands
        A sequence of commands to submit to the ledger. These commands are submitted atomically
        (in other words, they all succeed or they all fail).
    .. attribute:: CommandPayload.deduplication_time:
        The length of the time window during which all commands with the same party and command ID
        will be deduplicated. Duplicate commands submitted before the end of this window return an
        ``ALREADY_EXISTS`` error.
    """

    party: Party
    ledger_id: str
    workflow_id: str
    application_id: str
    command_id: str
    commands: "Sequence[api_types.Command]"
    deduplication_time: "Optional[timedelta]" = None

    def __post_init__(self):
        missing_fields = [
            field.name
            for field in fields(self)
            if field.name != "deduplication_time" and getattr(self, field.name) is None
        ]
        if missing_fields:
            raise ValueError(
                f"Some fields are set to None when they are required: " f"{missing_fields}"
            )


# noinspection PyDeprecation
def create(template, arguments=None):
    warnings.warn(
        "dazl.client.commands.create is deprecated; "
        "prefer calling dazl.ledger.Connection.create or "
        "dazl.client.PartyClient.submit_create, "
        "or use dazl.ledger.CreateCommand instead.",
        DeprecationWarning,
        stacklevel=2,
    )

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        return CreateCommand(template, arguments)


# noinspection PyDeprecation
def create_and_exercise(template, create_arguments, choice_name, choice_argument):
    warnings.warn(
        "dazl.client.commands.CreateAndExerciseCommand is deprecated; "
        "prefer calling dazl.ledger.Connection.create_and_exercise or "
        "dazl.client.PartyClient.submit_create_and_exercise, "
        "or use dazl.ledger.CreateAndExerciseCommand instead.",
        DeprecationWarning,
        stacklevel=2,
    )

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        return CreateAndExerciseCommand(template, create_arguments, choice_name, choice_argument)


# noinspection PyDeprecation
def exercise(contract, choice, arguments=None):
    warnings.warn(
        "dazl.client.commands.exercise is deprecated; "
        "prefer calling dazl.ledger.Connection.exercise or "
        "dazl.client.PartyClient.submit_exercise, "
        "or use dazl.ledger.ExerciseCommand instead.",
        DeprecationWarning,
        stacklevel=2,
    )

    if not isinstance(choice, str):
        raise ValueError(
            "choice must be a string name, a template type, or an instantiated template"
        )

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        return ExerciseCommand(contract, choice, arguments)


# noinspection PyDeprecation
def exercise_by_key(template, contract_key, choice_name, choice_argument):
    warnings.warn(
        "dazl.client.commands.ExerciseByKeyCommand is deprecated; "
        "prefer calling dazl.ledger.Connection.exercise_by_key or "
        "dazl.client.PartyClient.submit_exercise_by_key, "
        "or use dazl.ledger.ExerciseByKeyCommand instead.",
        DeprecationWarning,
        stacklevel=2,
    )

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        return ExerciseByKeyCommand(template, contract_key, choice_name, choice_argument)


EventHandlerResponse = Union[CommandsOrCommandSequence, CommandBuilder, CommandPayload]



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off

from typing import (
    AsyncIterable as __AsyncStream,
    Awaitable as __Awaitable,
    Iterable as __Stream,
    Union as __Union,
    overload,
)

from grpc import Channel as __Channel
from grpc.aio import Channel as __AsyncChannel

from .participant_pruning_service_pb2 import PruneRequest, PruneResponse

__all__ = [
    "ParticipantPruningServiceStub",
]

class ParticipantPruningServiceStub:
    @classmethod
    @overload
    def __new__(cls, channel: __Channel) -> _ParticipantPruningServiceStub: ...
    @classmethod
    @overload
    def __new__(cls, channel: __AsyncChannel) -> _ParticipantPruningServiceStub_Async: ...
    def Prune(self, __1: PruneRequest) -> __Union[PruneResponse, __Awaitable[PruneResponse]]: ...

class _ParticipantPruningServiceStub(ParticipantPruningServiceStub):
    def Prune(self, __1: PruneRequest) -> PruneResponse: ...

class _ParticipantPruningServiceStub_Async(ParticipantPruningServiceStub):
    def Prune(self, __1: PruneRequest) -> __Awaitable[PruneResponse]: ...



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
import sys
from typing import AbstractSet, Any, Iterator, Optional, Sequence, TypeVar, Union

from .. import (
    Connection as _Connection,
    PackageService as _PackageService,
    QueryStream as _QueryStream,
)
from ...damlast.daml_lf_1 import PackageRef, TypeConName
from ...prim import ContractData, ContractId
from ...query import Queries, Query
from ..api_types import ArchiveEvent, Boundary, Command, CreateEvent, ExerciseResponse, PartyInfo

if sys.version_info >= (3, 8):
    from typing import Protocol, runtime_checkable
else:
    from typing_extensions import Protocol, runtime_checkable

__all__ = ["PackageService", "Connection", "QueryStream"]

Self = TypeVar("Self")

class PackageService(_PackageService, Protocol):
    def get_package(self, package_id: PackageRef) -> bytes: ...
    def list_package_ids(self) -> AbstractSet[PackageRef]: ...

@runtime_checkable
class Connection(_Connection, PackageService, Protocol):
    def __enter__(self: Self) -> Self: ...
    def __exit__(self, exc_type, exc_val, exc_tb) -> None: ...
    def open(self) -> None: ...
    def close(self) -> None: ...
    def submit(
        self,
        __commands: Union[Command, Sequence[Command]],
        *,
        workflow_id: Optional[str] = None,
        command_id: Optional[str] = None,
    ) -> None: ...
    def create(
        self,
        __template_id: Union[str, TypeConName],
        __payload: ContractData,
        *,
        workflow_id: Optional[str] = None,
        command_id: Optional[str] = None,
    ) -> CreateEvent: ...
    def exercise(
        self,
        __contract_id: ContractId,
        __choice_name: str,
        __argument: Optional[ContractData] = None,
        *,
        workflow_id: Optional[str] = None,
        command_id: Optional[str] = None,
    ) -> ExerciseResponse: ...
    def create_and_exercise(
        self,
        __template_id: Union[str, TypeConName],
        __payload: ContractData,
        __choice_name: str,
        __argument: Optional[ContractData] = None,
        *,
        workflow_id: Optional[str] = None,
        command_id: Optional[str] = None,
    ) -> ExerciseResponse: ...
    def exercise_by_key(
        self,
        __template_id: Union[str, TypeConName],
        __choice_name: str,
        __key: Any,
        __argument: Optional[ContractData] = None,
        *,
        workflow_id: Optional[str] = None,
        command_id: Optional[str] = None,
    ) -> ExerciseResponse: ...
    def archive(
        self,
        __contract_id: ContractId,
        *,
        workflow_id: Optional[str] = None,
        command_id: Optional[str] = None,
    ) -> ArchiveEvent: ...
    def archive_by_key(
        self,
        __template_id: str,
        __key: Any,
        *,
        workflow_id: Optional[str] = None,
        command_id: Optional[str] = None,
    ) -> ArchiveEvent: ...
    def query(self, __template_id: str = "*", __query: Query = None) -> QueryStream: ...
    def query_many(self, *q: Queries) -> QueryStream: ...
    def stream(
        self, __template_id: str = "*", __query: Query = None, *, offset: Optional[str] = None
    ) -> QueryStream: ...
    def stream_many(self, *q: Queries, offset: Optional[str] = None) -> QueryStream: ...
    def allocate_party(
        self, *, identifier_hint: Optional[str] = None, display_name: Optional[str] = None
    ) -> PartyInfo: ...
    def list_known_parties(self) -> Sequence[PartyInfo]: ...
    def upload_package(self, contents: bytes) -> None: ...

@runtime_checkable
class QueryStream(_QueryStream, Protocol):
    def creates(self) -> Iterator[CreateEvent]: ...
    def events(self) -> Iterator[Union[CreateEvent, ArchiveEvent]]: ...
    def items(self) -> Iterator[Union[CreateEvent, ArchiveEvent, Boundary]]: ...
    def __iter__(self) -> Iterator[Union[CreateEvent, ArchiveEvent, Boundary]]: ...
    def run(self) -> None: ...
    def close(self) -> None: ...
    def __enter__(self: Self) -> Self: ...
    def __exit__(self, exc_type, exc_val, exc_tb) -> None: ...



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
DAML Types
----------

Constants and constructor functions for the various primitive DAML types.
"""
from typing import Mapping

from .daml_lf_1 import FieldWithType, PrimType, Type, TypeConName, TypeSynName

__all__ = [
    # Type.Var
    "var",
    # Type.Con
    "con",
    # Type.Prim
    "Unit",
    "Bool",
    "Int",
    "Decimal",
    "Text",
    "Time",
    "Party",
    "List",
    "Update",
    "Scenario",
    "Date",
    "ContractId",
    "Optional",
    "Arrow",
    "TextMap",
    "Numeric",
    "Any",
    "TypeRep",
    "Map",
    # Type.Struct
    "struct",
    # Type.ForAll (intentionally omitted)
    # Type.Syn
    "syn",
]


# region Variables (Type.Var)


def var(a: str, *args: "Type") -> "Type":
    """
    Construct a DAML :class:`Type` that is equivalent to ``a`` (a type variable of the specified
    name).
    """
    return Type(var=Type.Var(a, args=tuple(args)))


# endregion

# region Constructors (Type.Con)


def con(tycon: "TypeConName", *args: "Type") -> "Type":
    """
    Construct a DAML :class:`Type` that is a reference to a :class:`DefDataType` (a record, variant,
    or enum).

    :param tycon: The fully-qualified name of the type constructor.
    :param args: Type arguments (only applicable if the type is a generic type).
    """
    if tycon is None:
        raise ValueError("tycon is required")

    return Type(con=Type.Con(tycon=tycon, args=tuple(args)))


# endregion

# region Primitive types (Type.Prim)


def _prim_type(pt: "PrimType", *a: "Type") -> "Type":
    return Type(prim=Type.Prim(prim=pt, args=tuple(a)))


# DAML's unit type (``()``).
Unit = _prim_type(PrimType.UNIT)

# DAML's boolean type (``Bool``).
Bool = _prim_type(PrimType.BOOL)

# DAML's 64-bit integer type (``Int``).
Int = _prim_type(PrimType.INT64)

# DAML's historical decimal type (``Decimal``). You should generally use :func:`Numeric`(10)
# instead, which means the same thing.
Decimal = _prim_type(PrimType.DECIMAL)

# DAML's text type (``Text``).
Text = _prim_type(PrimType.TEXT)

# DAML's datetime type (``Time``).
Time = _prim_type(PrimType.TIMESTAMP)

# DAML's Party type (``Party``).
Party = _prim_type(PrimType.PARTY)


# noinspection PyPep8Naming
def List(a: "Type") -> "Type":
    """
    Construct a DAML :class:`Type` that is equivalent to ``[a]`` (a list of objects).
    """
    return Type(prim=Type.Prim(PrimType.LIST, args=(a,)))


# noinspection PyPep8Naming
def Update(a: "Type") -> "Type":
    """
    Construct a DAML :class:`Type` that is equivalent to ``Update a``.

    This type is not normally used in ``dazl`` since we are primarily concerned with serializable
    types; it is included for completeness.
    """
    return Type(prim=Type.Prim(PrimType.UPDATE, args=(a,)))


# noinspection PyPep8Naming
def Scenario(a: "Type") -> "Type":
    """
    Construct a DAML :class:`Type` that is equivalent to ``Scenario a``.

    This type is not normally used in ``dazl`` since we are primarily concerned with serializable
    types; it is included for completeness.
    """
    return Type(prim=Type.Prim(PrimType.SCENARIO, args=(a,)))


# DAML's Date type (``Date``).
Date = _prim_type(PrimType.DATE)


# noinspection PyPep8Naming
def ContractId(a: "Type") -> "Type":
    """
    Construct a DAML :class:`Type` that is equivalent to ``ContractId a``.

    No checking is done to verify that the argument refers to a valid template.
    """
    return Type(prim=Type.Prim(PrimType.CONTRACT_ID, args=(a,)))


# noinspection PyPep8Naming
def Optional(a: "Type") -> "Type":
    """
    Construct a DAML :class:`Type` that is equivalent to ``Optional a`` (either ``Some a`` or
    ``None``).
    """
    return Type(prim=Type.Prim(PrimType.OPTIONAL, args=(a,)))


# noinspection PyPep8Naming
def Arrow(a: "Type", b: "Type") -> "Type":
    """
    Construct a DAML :class:`Type` that is equivalent to ``a -> b`` (a function type).

    This type is not normally used in ``dazl`` since we are primarily concerned with serializable
    types; it is included for completeness.
    """
    return _prim_type(PrimType.ARROW, a, b)


# noinspection PyPep8Naming
def TextMap(a: "Type") -> "Type":
    """
    Construct a DAML :class:`Type` that is equivalent to ``TextMap a``.
    """
    return _prim_type(PrimType.TEXTMAP, a)


# noinspection PyPep8Naming
def Numeric(n: int) -> "Type":
    """
    Construct a DAML :class:`Type` that is equivalent to ``Numeric n``.
    """
    return _prim_type(PrimType.NUMERIC, Type(nat=n))


# DAML's DA.Internal.LF.Any type.
#
# This type is not normally used in ``dazl`` since we are primarily concerned with serializable
# types; it is included for completeness.
Any = _prim_type(PrimType.ANY)

# DAML's DA.Internal.LF.TypeRep type.
#
# This type is not normally used in ``dazl`` since we are primarily concerned with serializable
# types; it is included for completeness.
TypeRep = _prim_type(PrimType.TYPE_REP)


# noinspection PyPep8Naming
def Map(k: "Type", v: "Type") -> "Type":
    """
    Construct a DAML :class:`Type` that is equivalent to ``Map k v``.
    """
    return _prim_type(PrimType.GENMAP, k, v)


# endregion

# region Type.Forall

# A forall convenience constructor function is intentionally omitted.

# endregion

# region Type.Struct


def struct(fields: "Mapping[str, Type]"):
    """
    Construct a DAML :class:`Type` that is equivalent to a structural tuple type (``(a, b, ...)``).

    :param fields:
        A map of field names to their types. DAML-LF structs generally define structs as having
        field names ``"_1"``, ``"_2"``, ``"_3"``, and so on.
    """
    return Type(
        tuple=Type.Tuple(fields=tuple([FieldWithType(field, typ) for field, typ in fields.items()]))
    )


# endregion

# region Type.Syn


def syn(tysyn: "TypeSynName", *args: "Type") -> "Type":
    """
    Construct a DAML :class:`Type` that is a reference to another type.

    :param tysyn: The fully-qualified name of the type synonym.
    :param args: Type arguments (only applicable if the type is a generic type).
    """
    return Type(syn=Type.Syn(tysyn=tysyn, args=tuple(args)))


# endregion



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from enum import Enum


class OrderedEnum(Enum):
    """
    An enum that supports ordering semantics.
    (Taken from https://docs.python.org/3/library/enum.html#orderedenum)
    """

    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value
        return NotImplemented

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value
        return NotImplemented

    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.value <= other.value
        return NotImplemented

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Main module that exposes utility functions that can be called directly from the command line.
"""

from .cli import main

main()
